/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "alaudio/ALAudioBuffer.h"
#include "alaudio/ALAudioSource.h"
#include "alaudio/ALAudioSystem.h"
#include "convergence/AudioManager.h"
#include "jsonio/JSONParser.h"
#include "pcmaudio/PCMAudio.h"
#define OV_EXCLUDE_STATIC_CALLBACKS
#include "vorbisaudioio/VorbisAudioStream.h"
#include <stdio.h>
#include <string.h>

#define SOURCE_COUNT 16
#define MUSIC_BUFFER_COUNT 3
#define MUSIC_BUFFER_SIZE_SECONDS 1

static GameSession * gameSession;
static ALAudioSystem * audioSystem;
static ALAudioBuffer ** audioBuffers;
static ALAudioSource * audioSources[SOURCE_COUNT];
static struct JSONNode * soundList;
static int nextSourceIndex;
static ALAudioSource * musicSource;
static ALAudioBuffer * musicBuffers[MUSIC_BUFFER_COUNT];
static int musicBufferIndex;
static VorbisAudioStream * musicStream;
void * musicStreamBuffer;
static bool musicPlaying, musicPaused;

void AudioManager_globalInit(GameSession * inGameSession) {
	struct JSONParseError parseError;
	PCMAudio * audioData;
	unsigned int sourceIndex, bufferIndex;
	
	gameSession = inGameSession;
	audioSystem = ALAudioSystem_create();
	
	soundList = JSONParser_loadFile("sounds.json", &parseError);
	if (soundList == NULL) {
		fprintf(stderr, "Error: Couldn't parse sounds.json: %s (%zu)\n", parseError.description, parseError.charIndex);
		
	} else if (soundList->type != JSON_TYPE_OBJECT) {
		fprintf(stderr, "Error: sounds.json malformed (expected JSON_TYPE_OBJECT, but got %d)\n", soundList->type);
		
	} else {
		audioBuffers = calloc(sizeof(ALAudioBuffer *), soundList->value.count);
		for (bufferIndex = 0; bufferIndex < soundList->value.count; bufferIndex++) {
			if (soundList->subitems[bufferIndex].type != JSON_TYPE_STRING) {
				fprintf(stderr, "Error: sounds.json malformed (expected subitem %u to be JSON_TYPE_STRING, but got %d)\n", bufferIndex, soundList->subitems[bufferIndex].type);
				continue;
			}
			audioData = gameSession->resourceManager->referenceResource(gameSession->resourceManager, "vorbis", soundList->subitems[bufferIndex].value.string);
			if (audioData == NULL) {
				fprintf(stderr, "Error: Couldn't load \"%s\" (sounds.json subitem %u)\n", soundList->subitems[bufferIndex].value.string, bufferIndex);
				continue;
			}
			audioBuffers[bufferIndex] = ALAudioBuffer_create();
			audioBuffers[bufferIndex]->bufferPCMAudio(audioBuffers[bufferIndex], audioData);
			gameSession->resourceManager->releaseResource(gameSession->resourceManager, "vorbis", soundList->subitems[bufferIndex].value.string);
		}
	}
	
	for (sourceIndex = 0; sourceIndex < SOURCE_COUNT; sourceIndex++) {
		audioSources[sourceIndex] = ALAudioSource_create();
	}
	musicSource = ALAudioSource_create();
	for (bufferIndex = 0; bufferIndex < MUSIC_BUFFER_COUNT; bufferIndex++) {
		musicBuffers[bufferIndex] = ALAudioBuffer_create();
	}
}

void AudioManager_globalShutdown() {
	unsigned int sourceIndex, bufferIndex;
	
	for (sourceIndex = 0; sourceIndex < SOURCE_COUNT; sourceIndex++) {
		audioSources[sourceIndex]->dispose(audioSources[sourceIndex]);
	}
	musicSource->dispose(musicSource);
	if (soundList != NULL) {
		if (audioBuffers != NULL) {
			for (bufferIndex = 0; bufferIndex < soundList->value.count; bufferIndex++) {
				if (audioBuffers[bufferIndex] != NULL) {
					audioBuffers[bufferIndex]->dispose(audioBuffers[bufferIndex]);
				}
			}
			free(audioBuffers);
		}
		JSONNode_dispose(soundList);
	}
	for (bufferIndex = 0; bufferIndex < MUSIC_BUFFER_COUNT; bufferIndex++) {
		musicBuffers[bufferIndex]->dispose(musicBuffers[bufferIndex]);
	}
	audioSystem->dispose(audioSystem);
}

void AudioManager_play(const char * soundName) {
	if (soundList != NULL && gameSession->preferences->getBoolean(gameSession->preferences, "play_sound")) {
		size_t subitemIndex;
		
		subitemIndex = JSONNode_subitemIndexForKey(soundList, soundName, strlen(soundName));
		if (subitemIndex == JSON_SUBITEM_NOT_FOUND) {
			fprintf(stderr, "Warning: Request to play sound \"%s\", which wasn't specified in sounds.json\n", soundName);
			
		} else if (audioBuffers != NULL && audioBuffers[subitemIndex] != NULL) {
			audioSources[nextSourceIndex]->stop(audioSources[nextSourceIndex]);
			audioSources[nextSourceIndex]->load(audioSources[nextSourceIndex], audioBuffers[subitemIndex]);
			audioSources[nextSourceIndex]->play(audioSources[nextSourceIndex]);
			nextSourceIndex = (nextSourceIndex + 1) % SOURCE_COUNT;
		}
	}
}

static void bufferNextMusicChunk() {
	size_t bytesRead;
	
	bytesRead = musicStream->read(musicStream, MUSIC_BUFFER_SIZE_SECONDS * musicStream->bytesPerSample * musicStream->channelCount * musicStream->sampleRate, musicStreamBuffer, true);
	if (bytesRead < MUSIC_BUFFER_SIZE_SECONDS * musicStream->bytesPerSample * musicStream->channelCount * musicStream->sampleRate) {
		fprintf(stderr, "Error: Couldn't read %d bytes from music stream (got %zu); stopping music\n", MUSIC_BUFFER_SIZE_SECONDS * musicStream->bytesPerSample * musicStream->channelCount * musicStream->sampleRate, bytesRead);
		AudioManager_stopMusic();
		return;
	}
	
	musicBuffers[musicBufferIndex]->bufferAudioData(musicBuffers[musicBufferIndex],
	                                                musicStream->channelCount == 1 ? musicStream->bytesPerSample == 1 ? AL_FORMAT_MONO8 : AL_FORMAT_MONO16 : musicStream->bytesPerSample == 1 ? AL_FORMAT_STEREO8 : AL_FORMAT_STEREO16,
	                                                musicStreamBuffer,
	                                                bytesRead,
	                                                musicStream->sampleRate);
	alSourceQueueBuffers(musicSource->source, 1, &musicBuffers[musicBufferIndex]->buffer);
}

void AudioManager_startMusic(const char * musicName) {
	if (!gameSession->preferences->getBoolean(gameSession->preferences, "play_music")) {
		return;
	}
	
	if (musicStream != NULL) {
		musicStream->dispose(musicStream);
	}
	musicStream = VorbisAudioStream_createWithFile(musicName);
	
	musicSource->stop(musicSource);
	free(musicStreamBuffer);
	musicStreamBuffer = malloc(MUSIC_BUFFER_SIZE_SECONDS * musicStream->bytesPerSample * musicStream->channelCount * musicStream->sampleRate);
	
	for (musicBufferIndex = 0; musicBufferIndex < MUSIC_BUFFER_COUNT; musicBufferIndex++) {
		bufferNextMusicChunk();
	}
	musicBufferIndex = MUSIC_BUFFER_COUNT - 1;
	if (!musicPaused) {
		musicSource->play(musicSource);
	}
	musicPlaying = true;
}

void AudioManager_stopMusic() {
	if (musicPlaying) {
		int bufferIndex;
		
		musicSource->stop(musicSource);
		for (bufferIndex = 0; bufferIndex < MUSIC_BUFFER_COUNT; bufferIndex++) {
		  alSourceUnqueueBuffers(musicSource->source, 1, &musicBuffers[bufferIndex]->buffer);
		}
		musicStream->dispose(musicStream);
		musicStream = NULL;
		musicPlaying = false;
		musicPaused = false;
	}
}

void AudioManager_pauseMusic() {
	musicPaused = true;
	if (musicPlaying) {
		//musicSource->pause(musicSource);
		// HACK; replace with above when ALAudio pause bug fixed
		ALAudioSource_pause(musicSource);
	}
}

void AudioManager_resumeMusic() {
	musicPaused = false;
	if (musicPlaying) {
		musicSource->play(musicSource);
	}
}

void AudioManager_singleThreadStreamMusic() {
  int processed;
  
	if (!musicPlaying || musicPaused) {
		return;
	}
	
  alGetSourcei(musicSource->source, AL_BUFFERS_PROCESSED, &processed);
  if (processed < 1) {
    return;
  }
	musicBufferIndex = (musicBufferIndex + 1) % MUSIC_BUFFER_COUNT;
  alSourceUnqueueBuffers(musicSource->source, 1, &musicBuffers[musicBufferIndex]->buffer);
	bufferNextMusicChunk();
}
