/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/DungeonRoomModel.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

DungeonRoomModel * DungeonRoomModel_create() {
	stemobject_create_implementation(DungeonRoomModel, init)
}

void DungeonRoomModel_init(compat_type(DungeonRoomModel *) selfPtr) {
	DungeonRoomModel * self = selfPtr;
	
	StemObject_init(self);
	self->dispose = DungeonRoomModel_dispose;
	self->loadTileMap = DungeonRoomModel_loadTileMap;
	
	self->tileMapFileName = NULL;
	self->colorMapFileName = NULL;
	self->tileCanvasFileName = NULL;
	self->tileCountX = 0;
	self->tileCountY = 0;
	self->tiles = NULL;
	self->enemyCount = 0;
	self->enemies = NULL;
	self->objectCount = 0;
	self->objects = NULL;
	self->exitCount = 0;
	self->exits = NULL;
}

void DungeonRoomModel_dispose(compat_type(DungeonRoomModel *) selfPtr) {
	DungeonRoomModel * self = selfPtr;
	size_t enemyIndex;
	
	free(self->tileMapFileName);
	free(self->colorMapFileName);
	free(self->tileCanvasFileName);
	for (enemyIndex = 0; enemyIndex < self->enemyCount; enemyIndex++) {
		free(self->enemies[enemyIndex].patrolWaypoints);
	}
	free(self->tiles);
	free(self->enemies);
	free(self->objects);
	free(self->exits);
	StemObject_dispose(self);
}

DungeonRoomModel * DungeonRoomModel_deserialize(compat_type(DeserializationContext *) deserializationContext) {
	stemobject_deserialize_implementation(DungeonRoomModel, deserializationContext, loadSerializedData)
}

bool DungeonRoomModel_loadSerializedData(compat_type(DungeonRoomModel *) selfPtr, compat_type(DeserializationContext *) deserializationContext) {
	DungeonRoomModel * self = selfPtr;
	DeserializationContext * context = deserializationContext;
	uint16_t formatVersion;
	const char * name;
	const char * tileMapFileName, * colorMapFileName, * tileCanvasFileName;
	const char * exitString;
	size_t enemyCount, enemyIndex;
	struct DungeonEnemy * enemies = NULL;
	size_t objectCount, objectIndex;
	struct DungeonObject * objects = NULL;
	size_t exitCount, exitIndex;
	struct DungeonRoomExit * exits = NULL;
	
	context->beginStructure(context, "dungeon_room");
	formatVersion = context->readUInt16(context, "format_version");
	if (context->status == SERIALIZATION_ERROR_OK && formatVersion != DUNGEON_ROOM_SERIALIZATION_FORMAT_VERSION) {
		return false;
	}
	
	name = context->readString(context, "name");
	context->beginStructure(context, "tile_map");
	tileMapFileName = context->readString(context, "file");
	colorMapFileName = context->readString(context, "color_map");
	tileCanvasFileName = context->readString(context, "canvas");
	context->endStructure(context);
	enemyCount = context->beginArray(context, "enemies");
	enemies = calloc(sizeof(struct DungeonEnemy), enemyCount);
	for (enemyIndex = 0; enemyIndex < enemyCount; enemyIndex++) {
		context->beginStructure(context, NULL);
		enemies[enemyIndex].type = context->readEnumeration(context, "type",
			"berserker", DUNGEON_ENEMY_TYPE_BERSERKER,
			"ranger", DUNGEON_ENEMY_TYPE_RANGER,
			"first_boss", DUNGEON_ENEMY_TYPE_FIRST_BOSS,
		NULL);
		enemies[enemyIndex].rank = min(context->readUInt8(context, "rank"), 2);
		context->beginStructure(context, "start");
		context->beginStructure(context, "position");
		enemies[enemyIndex].startPosition.x = context->readFloat(context, "x");
		enemies[enemyIndex].startPosition.y = context->readFloat(context, "y");
		context->endStructure(context);
		enemies[enemyIndex].startFacing = context->readEnumeration(context, "facing", enumKV(NORTH), enumKV(NORTHEAST), enumKV(EAST), enumKV(SOUTHEAST), enumKV(SOUTH), enumKV(SOUTHWEST), enumKV(WEST), enumKV(NORTHWEST), NULL);
		context->endStructure(context);
		enemies[enemyIndex].inventoryItemCount = context->beginArray(context, "inventory");
		if (enemies[enemyIndex].inventoryItemCount > 0) {
			size_t inventoryItemIndex;
			
			enemies[enemyIndex].inventoryItems = malloc(sizeof(struct InventoryModel_item) * enemies[enemyIndex].inventoryItemCount);
			for (inventoryItemIndex = 0; inventoryItemIndex < enemies[enemyIndex].inventoryItemCount; inventoryItemIndex++) {
				context->beginStructure(context, NULL);
				enemies[enemyIndex].inventoryItems[inventoryItemIndex].type = context->readEnumeration(context, "type",
					"heal_powerup", INVENTORY_ITEM_TYPE_HEAL_POWERUP,
					"key", INVENTORY_ITEM_TYPE_KEY,
				NULL);
				switch (enemies[enemyIndex].inventoryItems[inventoryItemIndex].type) {
					case INVENTORY_ITEM_TYPE_HEAL_POWERUP:
						enemies[enemyIndex].inventoryItems[inventoryItemIndex].u.healPowerup.rank = min(context->readUInt16(context, "rank"), 2);
						break;
					case INVENTORY_ITEM_TYPE_KEY:
						enemies[enemyIndex].inventoryItems[inventoryItemIndex].u.key.keyID = context->readInt16(context, "key_id");
						break;
				}
				context->endStructure(context);
			}
		} else {
			enemies[enemyIndex].inventoryItems = NULL;
		}
		context->endArray(context);
		enemies[enemyIndex].patrolWaypointCount = context->beginArray(context, "patrol_route");
		if (enemies[enemyIndex].patrolWaypointCount > 0) {
			size_t waypointIndex;
			
			enemies[enemyIndex].patrolWaypoints = malloc(sizeof(Vector2) * enemies[enemyIndex].patrolWaypointCount);
			for (waypointIndex = 0; waypointIndex < enemies[enemyIndex].patrolWaypointCount; waypointIndex++) {
				context->beginStructure(context, NULL);
				enemies[enemyIndex].patrolWaypoints[waypointIndex].x = context->readFloat(context, "x");
				enemies[enemyIndex].patrolWaypoints[waypointIndex].y = context->readFloat(context, "y");
				context->endStructure(context);
			}
		} else {
			enemies[enemyIndex].patrolWaypoints = NULL;
		}
		context->endArray(context);
		context->endStructure(context);
	}
	context->endArray(context);
	objectCount = context->beginArray(context, "objects");
	objects = calloc(sizeof(struct DungeonObject), objectCount);
	for (objectIndex = 0; objectIndex < objectCount; objectIndex++) {
		context->beginStructure(context, NULL);
		objects[objectIndex].type = context->readEnumeration(context, "type",
			"heal_powerup", DUNGEON_OBJECT_TYPE_HEAL_POWERUP,
			"door", DUNGEON_OBJECT_TYPE_DOOR,
			"key", DUNGEON_OBJECT_TYPE_KEY,
			"rank_up_powerup", DUNGEON_OBJECT_TYPE_RANK_UP_POWERUP,
			"staircase", DUNGEON_OBJECT_TYPE_STAIRCASE,
		NULL);
		switch (objects[objectIndex].type) {
			case DUNGEON_OBJECT_TYPE_HEAL_POWERUP:
				context->beginStructure(context, "position");
				objects[objectIndex].u.healPowerup.position.x = context->readFloat(context, "x");
				objects[objectIndex].u.healPowerup.position.y = context->readFloat(context, "y");
				context->endStructure(context);
				objects[objectIndex].u.healPowerup.rank = min(context->readUInt16(context, "rank"), 2);
				break;
			case DUNGEON_OBJECT_TYPE_DOOR:
				objects[objectIndex].u.door.keyID = context->readInt16(context, "key_id");
				context->beginStructure(context, "bounds");
				objects[objectIndex].u.door.bounds.left = context->readFloat(context, "left");
				objects[objectIndex].u.door.bounds.right = context->readFloat(context, "right");
				objects[objectIndex].u.door.bounds.bottom = context->readFloat(context, "bottom");
				objects[objectIndex].u.door.bounds.top = context->readFloat(context, "top");
				context->endStructure(context);
				break;
			case DUNGEON_OBJECT_TYPE_KEY:
				objects[objectIndex].u.key.keyID = context->readInt16(context, "key_id");
				context->beginStructure(context, "position");
				objects[objectIndex].u.key.position.x = context->readFloat(context, "x");
				objects[objectIndex].u.key.position.y = context->readFloat(context, "y");
				context->endStructure(context);
				break;
			case DUNGEON_OBJECT_TYPE_RANK_UP_POWERUP:
				context->beginStructure(context, "position");
				objects[objectIndex].u.rankUpPowerup.position.x = context->readFloat(context, "x");
				objects[objectIndex].u.rankUpPowerup.position.y = context->readFloat(context, "y");
				context->endStructure(context);
				break;
			case DUNGEON_OBJECT_TYPE_STAIRCASE:
				objects[objectIndex].u.staircase.down = context->readEnumeration(context, "direction",
					"UP", false,
					"DOWN", true,
				NULL);
				context->beginStructure(context, "position");
				objects[objectIndex].u.staircase.position.x = context->readFloat(context, "x");
				objects[objectIndex].u.staircase.position.y = context->readFloat(context, "y");
				context->endStructure(context);
				break;
		}
		context->endStructure(context);
	}
	context->endArray(context);
	exitCount = context->beginDictionary(context, "exits");
	exits = malloc(sizeof(struct DungeonRoomExit) * exitCount);
	for (exitIndex = 0; exitIndex < exitCount; exitIndex++) {
		exitString = context->readNextDictionaryKey(context);
		if (exitString == NULL) {
			break;
		}
		exits[exitIndex].name = Atom_fromString(exitString);
		context->beginStructure(context, exitString);
		context->beginStructure(context, "bounds");
		exits[exitIndex].bounds.left = context->readFloat(context, "left");
		exits[exitIndex].bounds.right = context->readFloat(context, "right");
		exits[exitIndex].bounds.bottom = context->readFloat(context, "bottom");
		exits[exitIndex].bounds.top = context->readFloat(context, "top");
		context->endStructure(context);
		context->endStructure(context);
	}
	context->endDictionary(context);
	context->endStructure(context);
	
	if (context->status != SERIALIZATION_ERROR_OK) {
		for (enemyIndex = 0; enemyIndex < enemyCount; enemyIndex++) {
			free(enemies[enemyIndex].patrolWaypoints);
			free(enemies[enemyIndex].inventoryItems);
		}
		free(enemies);
		free(objects);
		free(exits);
		return false;
	}
	
	DungeonRoomModel_init(self);
	self->name = Atom_fromString(name);
	self->tileMapFileName = malloc(strlen(tileMapFileName) + 1);
	strcpy(self->tileMapFileName, tileMapFileName);
	self->colorMapFileName = malloc(strlen(colorMapFileName) + 1);
	strcpy(self->colorMapFileName, colorMapFileName);
	self->tileCanvasFileName = malloc(strlen(tileCanvasFileName) + 1);
	strcpy(self->tileCanvasFileName, tileCanvasFileName);
	self->enemyCount = enemyCount;
	self->enemies = enemies;
	self->objectCount = objectCount;
	self->objects = objects;
	self->exitCount = exitCount;
	self->exits = exits;
	
	return true;
}

void DungeonRoomModel_serialize(compat_type(DungeonRoomModel *) selfPtr, compat_type(SerializationContext *) serializationContext) {
	DungeonRoomModel * self = selfPtr;
	SerializationContext * context = serializationContext;
	size_t enemyIndex, waypointIndex, inventoryItemIndex, objectIndex, exitIndex;
	
	context->beginStructure(context, "dungeon");
	context->writeUInt16(context, "format_version", DUNGEON_ROOM_SERIALIZATION_FORMAT_VERSION);
	context->writeString(context, "name", self->name);
	context->beginStructure(context, "tile_map");
	context->writeString(context, "file", self->tileMapFileName);
	context->writeString(context, "color_map", self->colorMapFileName);
	context->writeString(context, "canvas", self->tileCanvasFileName);
	context->endStructure(context);
	context->beginArray(context, "enemies");
	for (enemyIndex = 0; enemyIndex < self->enemyCount; enemyIndex++) {
		context->beginStructure(context, NULL);
		context->writeEnumeration(context, "type", self->enemies[enemyIndex].type,
			"berserker", DUNGEON_ENEMY_TYPE_BERSERKER,
			"ranger", DUNGEON_ENEMY_TYPE_RANGER,
			"first_boss", DUNGEON_ENEMY_TYPE_FIRST_BOSS,
		NULL);
		context->writeUInt8(context, "rank", self->enemies[enemyIndex].rank);
		context->beginStructure(context, "start");
		context->beginStructure(context, "position");
		context->writeFloat(context, "x", self->enemies[enemyIndex].startPosition.x);
		context->writeFloat(context, "y", self->enemies[enemyIndex].startPosition.y);
		context->endStructure(context);
		context->writeEnumeration(context, "facing", self->enemies[enemyIndex].startFacing, enumKV(NORTH), enumKV(NORTHEAST), enumKV(EAST), enumKV(SOUTHEAST), enumKV(SOUTH), enumKV(SOUTHWEST), enumKV(WEST), enumKV(NORTHWEST), NULL);
		context->endStructure(context);
		context->beginArray(context, "inventory");
		for (inventoryItemIndex = 0; inventoryItemIndex < self->enemies[enemyIndex].inventoryItemCount; inventoryItemIndex++) {
			context->beginStructure(context, NULL);
			context->writeEnumeration(context, "type", self->enemies[enemyIndex].inventoryItems[inventoryItemIndex].type,
				"heal_powerup", INVENTORY_ITEM_TYPE_HEAL_POWERUP,
				"key", INVENTORY_ITEM_TYPE_KEY,
			NULL);
			switch (self->enemies[enemyIndex].inventoryItems[inventoryItemIndex].type) {
				case INVENTORY_ITEM_TYPE_HEAL_POWERUP:
					context->writeUInt16(context, "rank", self->enemies[enemyIndex].inventoryItems[inventoryItemIndex].u.healPowerup.rank);
					break;
				case INVENTORY_ITEM_TYPE_KEY:
					context->writeInt16(context, "key_id", self->enemies[enemyIndex].inventoryItems[inventoryItemIndex].u.key.keyID);
					break;
			}
			context->endStructure(context);
		}
		context->endArray(context);
		context->beginArray(context, "patrol_route");
		for (waypointIndex = 0; waypointIndex < self->enemies[enemyIndex].patrolWaypointCount; waypointIndex++) {
			context->beginStructure(context, NULL);
			context->writeFloat(context, "x", self->enemies[enemyIndex].patrolWaypoints[waypointIndex].x);
			context->writeFloat(context, "y", self->enemies[enemyIndex].patrolWaypoints[waypointIndex].y);
			context->endStructure(context);
		}
		context->endArray(context);
		context->endStructure(context);
	}
	context->endArray(context);
	context->beginArray(context, "objects");
	for (objectIndex = 0; objectIndex < self->objectCount; objectIndex++) {
		context->beginStructure(context, NULL);
		context->writeEnumeration(context, "type", self->objects[objectIndex].type,
			"heal_powerup", DUNGEON_OBJECT_TYPE_HEAL_POWERUP,
			"door", DUNGEON_OBJECT_TYPE_DOOR,
			"key", DUNGEON_OBJECT_TYPE_KEY,
			"rank_up_powerup", DUNGEON_OBJECT_TYPE_RANK_UP_POWERUP,
			"staircase", DUNGEON_OBJECT_TYPE_STAIRCASE,
			NULL);
		context->beginStructure(context, "position");
		context->endStructure(context);
		switch (self->objects[objectIndex].type) {
			case DUNGEON_OBJECT_TYPE_HEAL_POWERUP:
				context->beginStructure(context, "position");
				context->writeFloat(context, "x", self->objects[objectIndex].u.healPowerup.position.x);
				context->writeFloat(context, "y", self->objects[objectIndex].u.healPowerup.position.y);
				context->endStructure(context);
				context->writeUInt16(context, "rank", self->objects[objectIndex].u.healPowerup.rank);
				break;
			case DUNGEON_OBJECT_TYPE_DOOR:
				context->writeInt16(context, "key_id", self->objects[objectIndex].u.door.keyID);
				context->beginStructure(context, "position");
				context->writeFloat(context, "left", self->objects[objectIndex].u.door.bounds.left);
				context->writeFloat(context, "right", self->objects[objectIndex].u.door.bounds.right);
				context->writeFloat(context, "bottom", self->objects[objectIndex].u.door.bounds.bottom);
				context->writeFloat(context, "top", self->objects[objectIndex].u.door.bounds.top);
				context->endStructure(context);
				break;
			case DUNGEON_OBJECT_TYPE_KEY:
				context->writeInt16(context, "key_id", self->objects[objectIndex].u.key.keyID);
				context->beginStructure(context, "position");
				context->writeFloat(context, "x", self->objects[objectIndex].u.key.position.x);
				context->writeFloat(context, "y", self->objects[objectIndex].u.key.position.y);
				context->endStructure(context);
				break;
			case DUNGEON_OBJECT_TYPE_RANK_UP_POWERUP:
				context->beginStructure(context, "position");
				context->writeFloat(context, "x", self->objects[objectIndex].u.rankUpPowerup.position.x);
				context->writeFloat(context, "y", self->objects[objectIndex].u.rankUpPowerup.position.y);
				context->endStructure(context);
				break;
			case DUNGEON_OBJECT_TYPE_STAIRCASE:
				context->writeEnumeration(context, "direction", self->objects[objectIndex].u.staircase.down,
					"UP", false,
					"DOWN", true,
				NULL);
				context->beginStructure(context, "position");
				context->writeFloat(context, "x", self->objects[objectIndex].u.staircase.position.x);
				context->writeFloat(context, "y", self->objects[objectIndex].u.staircase.position.y);
				context->endStructure(context);
				break;
		}
		context->endStructure(context);
	}
	context->endArray(context);
	context->beginDictionary(context, "exits");
	for (exitIndex = 0; exitIndex < self->exitCount; exitIndex++) {
		context->beginStructure(context, self->exits[exitIndex].name);
		context->beginStructure(context, "bounds");
		context->writeFloat(context, "left", self->exits[exitIndex].bounds.left);
		context->writeFloat(context, "right", self->exits[exitIndex].bounds.right);
		context->writeFloat(context, "bottom", self->exits[exitIndex].bounds.bottom);
		context->writeFloat(context, "top", self->exits[exitIndex].bounds.top);
		context->endStructure(context);
		context->endStructure(context);
	}
	context->endDictionary(context);
	context->endStructure(context);
}

void DungeonRoomModel_loadTileMap(compat_type(DungeonRoomModel *) selfPtr, DungeonTileColorMap * colorMap, void * pixels, unsigned int width, unsigned int height, unsigned int bytesPerRow, unsigned int bytesPerPixel) {
	DungeonRoomModel * self = selfPtr;
	unsigned int rowIndex, columnIndex, colorIndex;
	unsigned int color;
	unsigned char * pixelData = pixels;
	
	self->tileCountX = width;
	self->tileCountY = height;
	self->tiles = calloc(sizeof(struct DungeonTile), self->tileCountX * self->tileCountY);
	for (rowIndex = 0; rowIndex < height; rowIndex++) {
		for (columnIndex = 0; columnIndex < width; columnIndex++) {
			color = (pixelData[rowIndex * bytesPerRow + columnIndex * bytesPerPixel] << 16) | (pixelData[rowIndex * bytesPerRow + columnIndex * bytesPerPixel + 1] << 8) | pixelData[rowIndex * bytesPerRow + columnIndex * bytesPerPixel + 2];
			for (colorIndex = 0; colorIndex < colorMap->colorCount; colorIndex++) {
				if (colorMap->colors[colorIndex].color == color) {
					self->tiles[rowIndex * self->tileCountX + columnIndex].type = colorMap->colors[colorIndex].type;
					self->tiles[rowIndex * self->tileCountX + columnIndex].canvasTileID = colorMap->colors[colorIndex].canvasTileID;
					break;
				}
			}
			if (colorIndex == colorMap->colorCount) {
				fprintf(stderr, "Warning: A dungeon tile map of dimensions %u, %u has unmapped color 0x%06X at %u, %u\n", width, height, color, columnIndex, rowIndex);
			}
		}
	}
}
