/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/EnemyModel.h"
#include "convergence/AIMath.h"
#include "convergence/DungeonKeyModel.h"
#include "convergence/HealPowerupModel.h"
#include "convergence/PlayerModel.h"
#include "convergence/PlayerSpeedsterModel.h"
#include "convergence/PlayerSpyModel.h"
#include "convergence/ProjectileModel.h"
#include "convergence/SwordSwingModel.h"
#include <math.h>
#include <string.h>

#define WAYPOINT_VISITED_DISTANCE 0.25f
#define PLAYER_PURSUIT_DISTANCE 0.25f
#define TRAVEL_PATH_SLOP_DISTANCE 0.0625f

static void sharedInit(EnemyModel * self) {
	self->dispose = EnemyModel_dispose;
	self->copy = EnemyModel_copy;
	self->update = EnemyModel_update;
	self->softReset = EnemyModel_softReset;
	self->protected_ivar(aiUpdate) = EnemyModel_aiUpdate;
	self->atRest = EnemyModel_atRest;
	self->wouldCollide = EnemyModel_wouldCollide;
	self->protected_ivar(collisionInternal) = EnemyModel_collisionInternal;
	self->setAIState = EnemyModel_setAIState;
	self->attack = EnemyModel_attack;
	self->protected_ivar(takeDamage) = EnemyModel_takeDamage;
	self->protected_ivar(moveTowardPoint) = EnemyModel_moveTowardPoint;
	
	self->type = GAME_OBJECT_TYPE_ENEMY;
	self->passiveVisionDistance = 7.0f;
	self->activeVisionDistance = 10.0f;
	self->stealthVisionDistance = 2.5f;
	self->acceleration = 0.001953125f;
	self->deceleration = 0.00390625f;
	self->velocityMax = 0.0234375f;
	self->knockbackResistance = 0.0f;
}

void setNextPatrolWaypointIndex(EnemyModel * self, size_t index) {
	self->nextPatrolWaypointIndex = index;
	free(self->travelPath.nodes);
	if (index < self->patrolWaypointCount) {
		int origin, destination;
		
		origin = floor(self->position.y) * self->stateModel->tileCountX + floor(self->position.x);
		destination = floor(self->patrolWaypoints[self->nextPatrolWaypointIndex].y) * self->stateModel->tileCountX + floor(self->patrolWaypoints[self->nextPatrolWaypointIndex].x);
		self->travelPath = AIBestPathToDestination(origin, destination, self->stateModel->tiles, self->stateModel->tileCountX, self->stateModel->tileCountY);
		self->nextTravelPathNodeIndex = 0;
	} else {
		self->travelPath.nodeCount = 0;
		self->travelPath.nodes = NULL;
	}
}

void EnemyModel_init(compat_type(EnemyModel *) selfPtr, DungeonRoomStateModel * stateModel, int rank, int health, Vector2 position, Direction facing, size_t inventoryItemCount, struct InventoryModel_item * inventoryItems, size_t patrolWaypointCount, Vector2 * patrolWaypoints) {
	EnemyModel * self = selfPtr;
	
	MovingEntityModel_init(self, stateModel);
	sharedInit(self);
	
	self->rank = rank;
	self->initialPosition = position;
	self->initialFacing = facing;
	self->inventory = InventoryModel_createWithItems(inventoryItemCount, inventoryItems);
	self->patrolWaypointCount = patrolWaypointCount;
	if (patrolWaypointCount > 0) {
		self->patrolWaypoints = malloc(sizeof(Vector2) * self->patrolWaypointCount);
		memcpy(self->patrolWaypoints, patrolWaypoints, sizeof(Vector2) * patrolWaypointCount);
	} else {
		self->patrolWaypoints = NULL;
	}
	self->maxHealth = health;
	self->health = health;
	self->position = position;
	self->facing = facing;
	self->immunity = 0;
	self->stun = 0;
	self->aiUpdateCooldown = ENEMY_AI_UPDATE_INTERVAL;
	self->canSeePlayer = false;
	self->aiState = ENEMY_AI_STATE_PATROL;
	self->hitReactCooldown = 0;
	self->lostGiveUpCooldown = 0;
	self->travelPath.nodeCount = 0;
	self->travelPath.nodes = NULL;
	if (self->patrolWaypointCount > 0 && fabs(self->patrolWaypoints[0].x - self->position.x) < WAYPOINT_VISITED_DISTANCE && fabs(self->patrolWaypoints[0].y - self->position.y) < WAYPOINT_VISITED_DISTANCE) {
		setNextPatrolWaypointIndex(self, 1);
	} else {
		setNextPatrolWaypointIndex(self, 0);
	}
}

void EnemyModel_initCopy(compat_type(EnemyModel *) selfPtr, compat_type(EnemyModel *) originalPtr) {
	EnemyModel * self = selfPtr, * original = originalPtr;
	
	MovingEntityModel_initCopy(self, original);
	sharedInit(self);
	
	self->rank = original->rank;
	self->initialPosition = original->initialPosition;
	self->initialFacing = original->initialFacing;
	self->inventory = original->inventory->copy(original->inventory);
	self->patrolWaypointCount = original->patrolWaypointCount;
	if (self->patrolWaypointCount > 0) {
		self->patrolWaypoints = malloc(sizeof(Vector2) * self->patrolWaypointCount);
		memcpy(self->patrolWaypoints, original->patrolWaypoints, sizeof(Vector2) * self->patrolWaypointCount);
	} else {
		self->patrolWaypoints = NULL;
	}
	self->nextPatrolWaypointIndex = original->nextPatrolWaypointIndex;
	self->maxHealth = original->maxHealth;
	self->health = original->health;
	self->knockbackResistance = original->knockbackResistance;
	self->immunity = original->immunity;
	self->lastHitBlocked = original->lastHitBlocked;
	self->stun = original->stun;
	self->aiUpdateCooldown = original->aiUpdateCooldown;
	self->canSeePlayer = original->canSeePlayer;
	self->lastKnownPlayerPosition = original->lastKnownPlayerPosition;
	self->travelPath.nodeCount = original->travelPath.nodeCount;
	if (self->travelPath.nodeCount > 0) {
		self->travelPath.nodes = malloc(sizeof(int) * self->travelPath.nodeCount);
		memcpy(self->travelPath.nodes, original->travelPath.nodes, sizeof(int) * self->travelPath.nodeCount);
	} else {
		self->travelPath.nodes = NULL;
	}
	self->nextTravelPathNodeIndex = original->nextTravelPathNodeIndex;
	self->passiveVisionDistance = original->passiveVisionDistance;
	self->activeVisionDistance = original->activeVisionDistance;
	self->stealthVisionDistance = original->stealthVisionDistance;
	self->aiState = original->aiState;
	self->hitReactCooldown = original->hitReactCooldown;
	self->lostGiveUpCooldown = original->lostGiveUpCooldown;
}

void EnemyModel_dispose(compat_type(EnemyModel *) selfPtr) {
	EnemyModel * self = selfPtr;
	
	self->inventory->dispose(self->inventory);
	free(self->patrolWaypoints);
	free(self->travelPath.nodes);
	StemObject_dispose(self);
}

compat_type(EnemyModel *) EnemyModel_copy(compat_type(EnemyModel *) selfPtr) {
	stemobject_copy_implementation(EnemyModel, initCopy)
}

void EnemyModel_update(compat_type(EnemyModel *) selfPtr) {
	EnemyModel * self = selfPtr;
	
	MovingEntityModel_update(self);
	
	if (self->aiState == ENEMY_AI_STATE_PATROL && self->patrolWaypointCount > 0) {
		if (self->travelPath.nodes != NULL && self->nextTravelPathNodeIndex < self->travelPath.nodeCount) {
			Vector2 nodePosition;
			
			nodePosition.x = self->travelPath.nodes[self->nextTravelPathNodeIndex] % self->stateModel->tileCountX + 0.5f;
			nodePosition.y = self->travelPath.nodes[self->nextTravelPathNodeIndex] / self->stateModel->tileCountX + 0.5f;
			self->protected_ivar(moveTowardPoint)(self, nodePosition, TRAVEL_PATH_SLOP_DISTANCE);
			if (fabs(nodePosition.x - self->position.x) < TRAVEL_PATH_SLOP_DISTANCE && fabs(nodePosition.y - self->position.y) < TRAVEL_PATH_SLOP_DISTANCE && self->nextTravelPathNodeIndex < self->travelPath.nodeCount) {
				self->nextTravelPathNodeIndex++;
			}
			if (self->nextTravelPathNodeIndex >= self->travelPath.nodeCount || (fabs(self->patrolWaypoints[self->nextPatrolWaypointIndex].x - self->position.x) < WAYPOINT_VISITED_DISTANCE && fabs(self->patrolWaypoints[self->nextPatrolWaypointIndex].y - self->position.y) < WAYPOINT_VISITED_DISTANCE)) {
				setNextPatrolWaypointIndex(self, (self->nextPatrolWaypointIndex + 1) % self->patrolWaypointCount);
			}
		}
	} else if (self->aiState == ENEMY_AI_STATE_PURSUE) {
		if (self->travelPath.nodes != NULL && self->nextTravelPathNodeIndex < self->travelPath.nodeCount) {
			Vector2 nodePosition;
			
			nodePosition.x = self->travelPath.nodes[self->nextTravelPathNodeIndex] % self->stateModel->tileCountX + 0.5f;
			nodePosition.y = self->travelPath.nodes[self->nextTravelPathNodeIndex] / self->stateModel->tileCountX + 0.5f;
			self->protected_ivar(moveTowardPoint)(self, nodePosition, TRAVEL_PATH_SLOP_DISTANCE);
			if (fabs(nodePosition.x - self->position.x) < TRAVEL_PATH_SLOP_DISTANCE && fabs(nodePosition.y - self->position.y) < TRAVEL_PATH_SLOP_DISTANCE && self->nextTravelPathNodeIndex < self->travelPath.nodeCount) {
				self->nextTravelPathNodeIndex++;
			}
		}
	}
	
	if (self->immunity > 0) {
		self->immunity--;
	}
	if (self->stun > 0) {
		self->stun--;
	}
	
	self->aiUpdateCooldown--;
	if (self->aiUpdateCooldown <= 0) {
		self->aiUpdateCooldown = ENEMY_AI_UPDATE_INTERVAL;
		self->protected_ivar(aiUpdate)(self);
	}
}

void EnemyModel_softReset(compat_type(EnemyModel *) selfPtr) {
	EnemyModel * self = selfPtr;
	
	self->position = self->initialPosition;
	self->facing = self->initialFacing;
	self->facingIntentX = self->facingIntentY = 0;
	self->velocity.x = 0.0f;
	self->velocity.y = 0.0f;
	self->health = self->maxHealth;
	self->immunity = 0;
	self->stun = 0;
	self->aiUpdateCooldown = ENEMY_AI_UPDATE_INTERVAL;
	self->canSeePlayer = false;
	self->setAIState(self, ENEMY_AI_STATE_PATROL);
	self->hitReactCooldown = 0;
	self->lostGiveUpCooldown = 0;
	if (self->patrolWaypointCount > 0 && fabs(self->patrolWaypoints[0].x - self->position.x) < WAYPOINT_VISITED_DISTANCE && fabs(self->patrolWaypoints[0].y - self->position.y) < WAYPOINT_VISITED_DISTANCE) {
		setNextPatrolWaypointIndex(self, 1);
	} else {
		setNextPatrolWaypointIndex(self, 0);
	}
}

void EnemyModel_aiUpdate(compat_type(EnemyModel *) selfPtr) {
	EnemyModel * self = selfPtr;
	PlayerModel * player;
	GameObjectModel ** objects;
	size_t objectCount, objectIndex;
	
	self->canSeePlayer = false;
	objects = AIObjectsInLineOfSight(self->position,
	                                 directionToVector(self->facing),
	                                 M_PI,
	                                 self->aiState == ENEMY_AI_STATE_PATROL ? self->passiveVisionDistance : self->activeVisionDistance,
	                                 self->stateModel->tiles,
	                                 self->stateModel->tileCountX,
	                                 self->stateModel->tileCountY,
	                                 self->stateModel->objects,
	                                 self->stateModel->objectCount,
	                                 &objectCount);
	for (objectIndex = 0; objectIndex < objectCount; objectIndex++) {
		if (objects[objectIndex]->type == GAME_OBJECT_TYPE_PLAYER) {
			int origin, destination;
			
			player = (PlayerModel *) objects[objectIndex];
			if (player->playerClass == PLAYER_CLASS_SPY && ((PlayerSpyModel *) player)->stealth &&
			    (player->position.x - self->position.x) * (player->position.x - self->position.x) + (player->position.y - self->position.y) * (player->position.y - self->position.y) > self->stealthVisionDistance * self->stealthVisionDistance) {
				continue;
			}
			self->canSeePlayer = true;
			self->lastKnownPlayerPosition = player->position;
			free(self->travelPath.nodes);
			origin = floor(self->position.y) * self->stateModel->tileCountX + floor(self->position.x);
			destination = floor(self->lastKnownPlayerPosition.y) * self->stateModel->tileCountX + floor(self->lastKnownPlayerPosition.x);
			self->travelPath = AIBestPathToDestination(origin, destination, self->stateModel->tiles, self->stateModel->tileCountX, self->stateModel->tileCountY);
			self->nextTravelPathNodeIndex = 0;
			player->spotted(player, self);
			break;
		}
	}
	free(objects);
	
	if (!self->canSeePlayer) {
		switch (self->aiState) {
			case ENEMY_AI_STATE_HIT_REACT:
				if (self->hitReactCooldown > 0) {
					self->hitReactCooldown -= ENEMY_AI_UPDATE_INTERVAL;
					if (self->hitReactCooldown <= 0) {
						self->setAIState(self, ENEMY_AI_STATE_PATROL);
					}
				}
				break;
				
			case ENEMY_AI_STATE_ATTACK:
				self->setAIState(self, ENEMY_AI_STATE_PURSUE);
			case ENEMY_AI_STATE_PURSUE:
				if (self->nextTravelPathNodeIndex >= self->travelPath.nodeCount ||
				    (fabs(self->lastKnownPlayerPosition.x - self->position.x) < PLAYER_PURSUIT_DISTANCE && fabs(self->lastKnownPlayerPosition.y - self->position.y) < PLAYER_PURSUIT_DISTANCE)) {
					self->lostGiveUpCooldown = 240;
					self->setAIState(self, ENEMY_AI_STATE_LOST);
				}
				break;
				
			case ENEMY_AI_STATE_LOST:
				self->lostGiveUpCooldown -= ENEMY_AI_UPDATE_INTERVAL;
				if (self->lostGiveUpCooldown <= 0) {
					self->setAIState(self, ENEMY_AI_STATE_PATROL);
					setNextPatrolWaypointIndex(self, self->nextPatrolWaypointIndex);
				}
				break;
				
			default:
				break;
		}
	}
}

bool EnemyModel_atRest(compat_type(EnemyModel *) selfPtr) {
	EnemyModel * self = selfPtr;
	
	return self->patrolWaypointCount == 0 &&
	       self->immunity <= 0 &&
	       self->stun <= 0 &&
	       (self->aiState != ENEMY_AI_STATE_HIT_REACT || self->hitReactCooldown <= 0) &&
	       MovingEntityModel_atRest(self);
}

bool EnemyModel_wouldCollide(compat_type(EnemyModel *) selfPtr, compat_type(GameObjectModel *) objectPtr) {
	EnemyModel * self = selfPtr;
	
	if (self->markedForRemoval) {
		return false;
	}
	switch (((GameObjectModel *) objectPtr)->type) {
		case GAME_OBJECT_TYPE_ENEMY:
		case GAME_OBJECT_TYPE_PLAYER:
		case GAME_OBJECT_TYPE_DOOR:
		case GAME_OBJECT_TYPE_ROOM_EXIT:
			return true;
			
		case GAME_OBJECT_TYPE_SWORD_SWING:
			return self->immunity <= 0 && !((SwordSwingModel *) objectPtr)->hostile;
			
		case GAME_OBJECT_TYPE_PROJECTILE:
			return self->immunity <= 0 && !((ProjectileModel *) objectPtr)->hostile;
			
		default:
			return false;
	}
}

void EnemyModel_collisionInternal(compat_type(EnemyModel *) selfPtr, compat_type(GameObjectModel *) objectPtr, const struct collision collision, float time) {
	EnemyModel * self = selfPtr;
	GameObjectModel * object = objectPtr;
	
	MovingEntityModel_collisionInternal(self, objectPtr, collision, time);
	if (object != NULL) {
		if (object->type == GAME_OBJECT_TYPE_SWORD_SWING) {
			SwordSwingModel * swing = objectPtr;
			
			self->velocity.x = collision.normal.x * swing->knockback * (1.0f - self->knockbackResistance);
			self->velocity.y = collision.normal.y * swing->knockback * (1.0f - self->knockbackResistance);
			self->protected_ivar(takeDamage)(self, swing->damage, vectorToNearestDirection(Vector2_init(-collision.normal.x, -collision.normal.y)));
			
		} else if (object->type == GAME_OBJECT_TYPE_PROJECTILE) {
			ProjectileModel * projectile = objectPtr;
			
			self->velocity.x = projectile->direction.x * projectile->knockback * (1.0f - self->knockbackResistance);
			self->velocity.y = projectile->direction.y * projectile->knockback * (1.0f - self->knockbackResistance);
			self->protected_ivar(takeDamage)(self, projectile->damage, (vectorToNearestDirection(projectile->direction) + 4) % 8);
			
		} else if (object->type == GAME_OBJECT_TYPE_PLAYER &&
		           ((PlayerModel *) object)->playerClass == PLAYER_CLASS_SPEEDSTER &&
		           ((PlayerSpeedsterModel *) object)->dashDamage > 0) {
			PlayerSpeedsterModel * speedster = (PlayerSpeedsterModel *) object;
			float dot;
			
			dot = Vector2_dot(collision.normal, self->velocity);
			self->velocity.x -= collision.normal.x * (dot + speedster->dashKnockback) * (1.0f - self->knockbackResistance);
			self->velocity.y -= collision.normal.y * (dot + speedster->dashKnockback) * (1.0f - self->knockbackResistance);
			self->protected_ivar(takeDamage)(self, speedster->dashDamage, vectorToNearestDirection(Vector2_init(collision.normal.x, collision.normal.y)));
		}
	}
}

void EnemyModel_setAIState(compat_type(EnemyModel *) selfPtr, enum EnemyAIState state) {
	EnemyModel * self = selfPtr;
	
	if (self->aiState != state) {
		self->stopMoving(self);
		self->aiState = state;
	}
}

void EnemyModel_attack(compat_type(EnemyModel *) selfPtr) {
}

void EnemyModel_takeDamage(compat_type(EnemyModel *) selfPtr, int amount, Direction fromDirection) {
	EnemyModel * self = selfPtr;
	
	self->immunity = 15;
	self->lastHitBlocked = amount == 0;
	self->stun = 20;
	self->health -= amount;
	self->facingIntentX = self->facingIntentY = 0;
	self->facing = fromDirection;
	if (self->aiState == ENEMY_AI_STATE_PATROL || self->aiState == ENEMY_AI_STATE_LOST) {
		self->hitReactCooldown = 100;
		self->setAIState(self, ENEMY_AI_STATE_HIT_REACT);
	}
	if (self->health <= 0) {
		size_t itemIndex;
		compat_type(GameObjectModel *) item;
		
		for (itemIndex = 0; itemIndex < self->inventory->itemCount; itemIndex++) {
			switch (self->inventory->items[itemIndex].type) {
				case INVENTORY_ITEM_TYPE_HEAL_POWERUP:
					item = HealPowerupModel_create(self->stateModel, self->inventory->items[itemIndex].u.healPowerup.rank, self->position);
					break;
				case INVENTORY_ITEM_TYPE_KEY:
					item = DungeonKeyModel_create(self->stateModel, self->inventory->items[itemIndex].u.key.keyID, self->position);
					break;
				default:
					continue;
			}
			self->stateModel->addObject(self->stateModel, item);
		}
		self->markedForRemoval = true;
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), NULL);
	} else {
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(ENEMY_EVENT_HIT), &amount);
	}
}

void EnemyModel_moveTowardPoint(compat_type(EnemyModel *) selfPtr, Vector2 point, float slop) {
	EnemyModel * self = selfPtr;
	
	if (point.x < self->position.x - slop) {
		self->startMovingLeft(self);
	} else if (point.x > self->position.x + slop) {
		self->startMovingRight(self);
	} else {
		self->stopMovingLeft(self);
		self->stopMovingRight(self);
	}
	if (point.y < self->position.y - slop) {
		self->startMovingDown(self);
	} else if (point.y > self->position.y + slop) {
		self->startMovingUp(self);
	} else {
		self->stopMovingDown(self);
		self->stopMovingUp(self);
	}
}
