/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/EnemyRangerModel.h"
#include "convergence/ProjectileModel.h"
#include <stdlib.h>

#define ATTACK_DISTANCE_RANK_0 7.0f
#define ATTACK_DISTANCE_RANK_1 7.5f
#define ATTACK_DISTANCE_RANK_2 8.5f
#define SHOTS_BEFORE_REPOSITION_RANK_0 2
#define SHOTS_BEFORE_REPOSITION_RANK_1 4
#define SHOTS_BEFORE_REPOSITION_RANK_2 6
#define REPOSITION_UPDATES_RANK_0 110
#define REPOSITION_UPDATES_RANK_1 100
#define REPOSITION_UPDATES_RANK_2 80

EnemyRangerModel * EnemyRangerModel_create(DungeonRoomStateModel * stateModel, int rank, Vector2 position, Direction facing, size_t inventoryItemCount, struct InventoryModel_item * inventoryItems, size_t patrolWaypointCount, Vector2 * patrolWaypoints) {
	stemobject_create_implementation(EnemyRangerModel, init, stateModel, rank, position, facing, inventoryItemCount, inventoryItems, patrolWaypointCount, patrolWaypoints)
}

static void sharedInit(EnemyRangerModel * self) {
	self->dispose = EnemyRangerModel_dispose;
	self->copy = EnemyRangerModel_copy;
	self->update = EnemyRangerModel_update;
	self->protected_ivar(aiUpdate) = EnemyRangerModel_aiUpdate;
	self->atRest = EnemyRangerModel_atRest;
	self->setAIState = EnemyRangerModel_setAIState;
	self->attack = EnemyRangerModel_attack;
	
	self->enemyType = ENEMY_TYPE_RANGER;
	self->passiveVisionDistance = 9.0f;
	self->acceleration = 1.0f / 512.0f;
	self->deceleration = 2.0f / 512.0f;
	self->velocityMax = 12.0f / 512.0f;
}

void EnemyRangerModel_init(compat_type(EnemyRangerModel *) selfPtr, DungeonRoomStateModel * stateModel, int rank, Vector2 position, Direction facing, size_t inventoryItemCount, struct InventoryModel_item * inventoryItems, size_t patrolWaypointCount, Vector2 * patrolWaypoints) {
	EnemyRangerModel * self = selfPtr;
	int health;
	
	switch (rank) {
		case 0:
		default:
			health = 3;
			self->shotsBeforeReposition = SHOTS_BEFORE_REPOSITION_RANK_0;
			break;
		case 1:
			health = 12;
			self->shotsBeforeReposition = SHOTS_BEFORE_REPOSITION_RANK_1;
			break;
		case 2:
			health = 30;
			self->shotsBeforeReposition = SHOTS_BEFORE_REPOSITION_RANK_2;
			break;
	}
	EnemyModel_init(self, stateModel, rank, health, position, facing, inventoryItemCount, inventoryItems, patrolWaypointCount, patrolWaypoints);
	sharedInit(self);
	self->attackCooldown = 0;
}

void EnemyRangerModel_initCopy(compat_type(EnemyRangerModel *) selfPtr, compat_type(EnemyRangerModel *) originalPtr) {
	EnemyRangerModel * self = selfPtr, * original = originalPtr;
	
	EnemyModel_initCopy(self, original);
	sharedInit(self);
	
	self->attackCooldown = original->attackCooldown;
	self->shotsBeforeReposition = original->shotsBeforeReposition;
	self->repositionUpdatesRemaining = original->repositionUpdatesRemaining;
}

void EnemyRangerModel_dispose(compat_type(EnemyRangerModel *) selfPtr) {
	EnemyRangerModel * self = selfPtr;
	
	EnemyModel_dispose(self);
}

compat_type(EnemyRangerModel *) EnemyRangerModel_copy(compat_type(EnemyRangerModel *) selfPtr) {
	stemobject_copy_implementation(EnemyRangerModel, initCopy)
}

void EnemyRangerModel_update(compat_type(EnemyRangerModel *) selfPtr) {
	EnemyRangerModel * self = selfPtr;
	
	EnemyModel_update(self);
	
	if (self->aiState == ENEMY_AI_STATE_ATTACK) {
		self->facing = vectorToNearestDirection(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
		if (self->repositionUpdatesRemaining > 0) {
			Vector2 desiredPlayerOffset;
			
			desiredPlayerOffset = directionToVector(vectorToNearestDirection(Vector2_subtract(self->position, self->lastKnownPlayerPosition)));
			desiredPlayerOffset.x *= 4.0f;
			desiredPlayerOffset.y *= 4.0f;
			self->protected_ivar(moveTowardPoint)(self, Vector2_add(self->lastKnownPlayerPosition, desiredPlayerOffset), 0.25f);
			self->repositionUpdatesRemaining--;
			if (self->repositionUpdatesRemaining <= 0) {
				self->stopMoving(self);
				switch (self->rank) {
					case 0:
					default:
						self->shotsBeforeReposition = SHOTS_BEFORE_REPOSITION_RANK_0;
						break;
					case 1:
						self->shotsBeforeReposition = SHOTS_BEFORE_REPOSITION_RANK_1;
						break;
					case 2:
						self->shotsBeforeReposition = SHOTS_BEFORE_REPOSITION_RANK_2;
						break;
				}
			}
		} else {
			self->attack(self);
			if (self->shotsBeforeReposition <= 0) {
				switch (self->rank) {
					case 0:
					default:
						self->repositionUpdatesRemaining = REPOSITION_UPDATES_RANK_0;
						break;
					case 1:
						self->repositionUpdatesRemaining = REPOSITION_UPDATES_RANK_1;
						break;
					case 2:
						self->repositionUpdatesRemaining = REPOSITION_UPDATES_RANK_2;
						break;
				}
			}
		}
		self->facing = vectorToNearestDirection(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
	}
	
	if (self->attackCooldown > 0) {
		self->attackCooldown--;
	}
}

void EnemyRangerModel_aiUpdate(compat_type(EnemyRangerModel *) selfPtr) {
	EnemyRangerModel * self = selfPtr;
	
	if (self->aiState == ENEMY_AI_STATE_ATTACK) {
		self->facing = vectorToNearestDirection(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
	}
	EnemyModel_aiUpdate(self);
	if (self->canSeePlayer) {
		float attackDistance;
		
		switch (self->rank) {
			case 0:
			default:
				attackDistance = ATTACK_DISTANCE_RANK_0;
				break;
			case 1:
				attackDistance = ATTACK_DISTANCE_RANK_1;
				break;
			case 2:
				attackDistance = ATTACK_DISTANCE_RANK_2;
				break;
		}
		if ((self->position.x - self->lastKnownPlayerPosition.x) * (self->position.x - self->lastKnownPlayerPosition.x) + (self->position.y - self->lastKnownPlayerPosition.y) * (self->position.y - self->lastKnownPlayerPosition.y) < attackDistance * attackDistance) {
			self->setAIState(self, ENEMY_AI_STATE_ATTACK);
		} else {
			self->setAIState(self, ENEMY_AI_STATE_PURSUE);
		}
	}
}

bool EnemyRangerModel_atRest(compat_type(EnemyRangerModel *) selfPtr) {
	EnemyRangerModel * self = selfPtr;
	
	return self->attackCooldown <= 0 &&
	       EnemyModel_atRest(self);
}

void EnemyRangerModel_setAIState(compat_type(EnemyRangerModel *) selfPtr, enum EnemyAIState state) {
	EnemyRangerModel * self = selfPtr;
	
	if (state != self->aiState) {
		EnemyModel_setAIState(self, state);
		switch (state) {
			case ENEMY_AI_STATE_PURSUE:
			case ENEMY_AI_STATE_ATTACK:
				self->acceleration = 2.0f / 512.0f;
				self->deceleration = 4.0f / 512.0f;
				switch (self->rank) {
					case 0:
					default:
						self->velocityMax = 17.0f / 512.0f;
						break;
					case 1:
						self->velocityMax = 19.0f / 512.0f;
						break;
					case 2:
						self->velocityMax = 22.0f / 512.0f;
						break;
				}
				break;
				
			case ENEMY_AI_STATE_PATROL:
			default:
				self->acceleration = 1.0f / 512.0f;
				self->deceleration = 2.0f / 512.0f;
				self->velocityMax = 12.0f / 512.0f;
				break;
		}
		if (state == ENEMY_AI_STATE_ATTACK) {
			self->repositionUpdatesRemaining = 0;
		}
	}
}

void EnemyRangerModel_attack(compat_type(EnemyRangerModel *) selfPtr) {
	EnemyRangerModel * self = selfPtr;
	
	if (self->attackCooldown <= 0 && self->stun <= 0) {
		ProjectileModel * projectile;
		
		self->shotsBeforeReposition--;
		switch (self->rank) {
			case 0:
			default:
				self->attackCooldown = 35;
				projectile = ProjectileModel_create(self->stateModel, self->position, directionToVector(self->facing), 0.09375f, 7.0f, true, 1, 0.0625f);
				break;
			case 1:
				self->attackCooldown = 25;
				projectile = ProjectileModel_create(self->stateModel, self->position, directionToVector(self->facing), 0.09375f, 8.0f, true, 2, 0.0625f);
				break;
			case 2:
				self->attackCooldown = 15;
				projectile = ProjectileModel_create(self->stateModel, self->position, directionToVector(self->facing), 0.125f, 12.0f, true, (self->shotsBeforeReposition % 3) == 0 ? 5 : 4, 0.09375f);
				break;
		}
		projectile->position.x += sin(directionToRadians(self->facing)) * (self->radius - projectile->radius);
		projectile->position.y += cos(directionToRadians(self->facing)) * (self->radius - projectile->radius);
		self->stateModel->addObject(self->stateModel, projectile);
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(ENEMY_EVENT_SHOOT), NULL);
	}
}
