/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/AudioManager.h"
#include "convergence/EnemyFirstBossModel.h"
#include "convergence/EnemyModel.h"
#include "convergence/EnemyView.h"
#include <stddef.h>
#include <stdlib.h>

struct vertex_p2f_c3f {
	GLfloat position[2];
	GLfloat color[3];
};

EnemyView * EnemyView_create(compat_type(EnemyModel *) modelPtr) {
	stemobject_create_implementation(EnemyView, init, modelPtr);
}

static bool hit(void * sender, Atom eventID, void * eventData, void * context) {
	if (*(int *) eventData == 0) {
		AudioManager_play("enemy_block");
	} else {
		AudioManager_play("enemy_hit");
	}
	return true;
}

static bool soundEffect(void * sender, Atom eventID, void * eventData, void * context) {
	AudioManager_play(context);
	return true;
}

static char * swordSwingConst = "sword_swing";
static char * shootConst = "shoot";
static char * enemyDeathConst = "enemy_death";

void EnemyView_init(compat_type(EnemyView *) selfPtr, compat_type(EnemyModel *) modelPtr) {
	EnemyView * self = selfPtr;
	struct vertex_p2f_c3f vertices[8] = {
		{{-0.375f, -0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{-0.375f,  0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{ 0.0f,   -0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{ 0.0f,    0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{ 0.0f,   -0.03125f}, {0.0f, 0.0f, 0.0f}},
		{{ 0.0f,    0.03125f}, {0.0f, 0.0f, 0.0f}},
		{{ 0.375f, -0.03125f}, {0.0f, 0.0f, 0.0f}},
		{{ 0.375f,  0.03125f}, {0.0f, 0.0f, 0.0f}}
	};
	GLushort indexes[12] = {
		0, 2, 1,
		2, 3, 1,
		4, 6, 5,
		6, 7, 5
	};
	
	MovingEntityView_init(self, modelPtr);
	self->dispose = EnemyView_dispose;
	self->draw = EnemyView_draw;
	
	glGenBuffers(1, &self->healthVBOID);
	glGenBuffers(1, &self->healthEBOID);
	glBindBuffer(GL_ARRAY_BUFFER, self->healthVBOID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->healthEBOID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_c3f) * 8, vertices, GL_STREAM_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * 12, indexes, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_ATTACK), soundEffect, swordSwingConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_SHOOT), soundEffect, shootConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_HIT), hit, self);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), soundEffect, enemyDeathConst);
}

void EnemyView_dispose(compat_type(EnemyView *) selfPtr) {
	EnemyView * self = selfPtr;
	
	glDeleteBuffers(1, &self->healthVBOID);
	glDeleteBuffers(1, &self->healthEBOID);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_ATTACK), soundEffect, swordSwingConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_SHOOT), soundEffect, shootConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_HIT), hit, self);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), soundEffect, enemyDeathConst);
	MovingEntityView_dispose(selfPtr);
}

void EnemyView_draw(compat_type(EnemyView *) selfPtr, Matrix modelviewMatrix, double deltaTime) {
	EnemyView * self = selfPtr;
	EnemyModel * model = (EnemyModel *) self->model;
	struct vertex_p2f_c3f vertices[4];
	float centerX;
	
	if (model->immunity > 0 && !model->lastHitBlocked) {
		self->outerColor = 0xFFFFFFFF;
		self->innerColor = 0xDFCF7FFF;
	} else {
		switch (model->enemyType) {
			case ENEMY_TYPE_BERSERKER:
				switch (model->rank) {
					case 0:
					default:
						self->outerColor = 0x974515FF;
						self->innerColor = 0xA74E16FF;
						break;
					case 1:
						self->outerColor = 0x7D0D0EFF;
						self->innerColor = 0x8D0D0EFF;
						break;
					case 2:
						self->outerColor = 0x7C005AFF;
						self->innerColor = 0x8C0072FF;
						break;
				}
				break;
			case ENEMY_TYPE_RANGER:
				switch (model->rank) {
					case 0:
					default:
						self->outerColor = 0x005F00FF;
						self->innerColor = 0x006F00FF;
						break;
					case 1:
						self->outerColor = 0x164D3FFF;
						self->innerColor = 0x1B5E3FFF;
						break;
					case 2:
						self->outerColor = 0x0B2E4EFF;
						self->innerColor = 0x0E3860FF;
						break;
				}
				break;
			case ENEMY_TYPE_FIRST_BOSS:
				if (((EnemyFirstBossModel *) model)->shotsLeft <= 0) {
					self->outerColor = 0xEFEFAFFF;
					self->innerColor = 0xDFDF8FFF;
				} else {
					switch (model->rank) {
						case 0:
						default:
							self->outerColor = 0x789E00FF;
							self->innerColor = 0x91BA18FF;
							break;
						case 1:
							self->outerColor = 0x9AA26FFF;
							self->innerColor = 0xB7BE92FF;
							break;
						case 2:
							self->outerColor = 0x000000FF;
							self->innerColor = 0x3F3F3FFF;
							break;
					}
				}
				break;
		}
	}
	MovingEntityView_draw(self, modelviewMatrix, deltaTime);
	
	centerX = -0.375f + 0.75f * model->health / model->maxHealth;
	vertices[0].position[0] = centerX;
	vertices[0].position[1] = -0.03125f;
	vertices[0].color[0] = 1.0f;
	vertices[0].color[1] = 1.0f;
	vertices[0].color[2] = 1.0f;
	vertices[1].position[0] = centerX;
	vertices[1].position[1] = 0.03125f;
	vertices[1].color[0] = 1.0f;
	vertices[1].color[1] = 1.0f;
	vertices[1].color[2] = 1.0f;
	vertices[2].position[0] = centerX;
	vertices[2].position[1] = -0.03125f;
	vertices[2].color[0] = 0.0f;
	vertices[2].color[1] = 0.0f;
	vertices[2].color[2] = 0.0f;
	vertices[3].position[0] = centerX;
	vertices[3].position[1] = 0.03125f;
	vertices[3].color[0] = 0.0f;
	vertices[3].color[1] = 0.0f;
	vertices[3].color[2] = 0.0f;
	
	glBindBuffer(GL_ARRAY_BUFFER, self->healthVBOID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->healthEBOID);
	glBufferSubData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_c3f) * 2, sizeof(struct vertex_p2f_c3f) * 4, vertices);
	
	Matrix_translate(&modelviewMatrix, model->position.x, model->position.y + 0.75f, 0.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_c3f), (void *) offsetof(struct vertex_p2f_c3f, position));
	glColorPointer(3, GL_FLOAT, sizeof(struct vertex_p2f_c3f), (void *) offsetof(struct vertex_p2f_c3f, color));
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glDrawElements(GL_TRIANGLES, 12, GL_UNSIGNED_SHORT, 0);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
