/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/InputController.h"
#include "gamepad/Gamepad.h"
#include "shell/ShellKeyCodes.h"

#define DEVICE_POLL_INTERVAL 120

InputController * InputController_create(PlayerInputMap * inputMap) {
	stemobject_create_implementation(InputController, init, inputMap)
}

static bool gamepadButtonDown(void * sender, Atom eventID, void * eventData, void * context) {
	InputController * self = context;
	struct Gamepad_buttonEvent * event = eventData;
	size_t gamepadIndex, bindingIndex;
	bool handled = false;
	
	for (gamepadIndex = 0; gamepadIndex < self->inputMap->gamepadMapCount; gamepadIndex++) {
		if (self->inputMap->gamepadMaps[gamepadIndex].vendorID == event->device->vendorID && self->inputMap->gamepadMaps[gamepadIndex].productID == event->device->productID) {
			for (bindingIndex = 0; bindingIndex < self->inputMap->gamepadMaps[gamepadIndex].buttonBindingCount; bindingIndex++) {
				if (self->inputMap->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].buttonID == event->buttonID) {
					handled = self->protected_ivar(triggerAction)(self, self->inputMap->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].actionID) || handled;
				}
			}
			break;
		}
	}
	return handled;
}

static bool gamepadButtonUp(void * sender, Atom eventID, void * eventData, void * context) {
	InputController * self = context;
	struct Gamepad_buttonEvent * event = eventData;
	size_t gamepadIndex, bindingIndex;
	bool handled = false;
	
	for (gamepadIndex = 0; gamepadIndex < self->inputMap->gamepadMapCount; gamepadIndex++) {
		if (self->inputMap->gamepadMaps[gamepadIndex].vendorID == event->device->vendorID && self->inputMap->gamepadMaps[gamepadIndex].productID == event->device->productID) {
			for (bindingIndex = 0; bindingIndex < self->inputMap->gamepadMaps[gamepadIndex].buttonBindingCount; bindingIndex++) {
				if (self->inputMap->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].buttonID == event->buttonID) {
					handled = self->protected_ivar(releaseAction)(self, self->inputMap->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].actionID) || handled;
				}
			}
			break;
		}
	}
	return handled;
}

static bool gamepadAxisMoved(void * sender, Atom eventID, void * eventData, void * context) {
	InputController * self = context;
	struct Gamepad_axisEvent * event = eventData;
	size_t virtualButtonIndex;
	bool handled = false;
	
	for (virtualButtonIndex = 0; virtualButtonIndex < self->virtualButtonCount; virtualButtonIndex++) {
		if (self->virtualButtons[virtualButtonIndex]->device == event->device && self->virtualButtons[virtualButtonIndex]->axisID == event->axisID) {
			self->virtualButtons[virtualButtonIndex]->update(self->virtualButtons[virtualButtonIndex], event->value);
			handled = true;
		}
	}
	return handled;
}

static bool virtualButtonDown(void * sender, Atom eventID, void * eventData, void * context) {
	GamepadAxisVirtualButton * virtualButton = sender;
	InputController * self = context;
	
	return self->protected_ivar(triggerAction)(self, virtualButton->actionID);
}

static bool virtualButtonUp(void * sender, Atom eventID, void * eventData, void * context) {
	GamepadAxisVirtualButton * virtualButton = sender;
	InputController * self = context;
	
	return self->protected_ivar(releaseAction)(self, virtualButton->actionID);
}

static bool deviceAttached(void * sender, Atom eventID, void * eventData, void * context) {
	InputController * self = context;
	struct Gamepad_device * device = eventData;
	size_t gamepadIndex;
	
	device->eventDispatcher->registerForEvent(device->eventDispatcher, Atom_fromString(GAMEPAD_EVENT_BUTTON_DOWN), gamepadButtonDown, self);
	device->eventDispatcher->registerForEvent(device->eventDispatcher, Atom_fromString(GAMEPAD_EVENT_BUTTON_UP), gamepadButtonUp, self);
	device->eventDispatcher->registerForEvent(device->eventDispatcher, Atom_fromString(GAMEPAD_EVENT_AXIS_MOVED), gamepadAxisMoved, self);
	for (gamepadIndex = 0; gamepadIndex < self->inputMap->gamepadMapCount; gamepadIndex++) {
		if (self->inputMap->gamepadMaps[gamepadIndex].vendorID == device->vendorID && self->inputMap->gamepadMaps[gamepadIndex].productID == device->productID) {
			size_t bindingIndex;
			
			self->virtualButtons = realloc(self->virtualButtons, sizeof(GamepadAxisVirtualButton *) * (self->virtualButtonCount + self->inputMap->gamepadMaps[gamepadIndex].axisBindingCount));
			for (bindingIndex = 0; bindingIndex < self->inputMap->gamepadMaps[gamepadIndex].axisBindingCount; bindingIndex++) {
				self->virtualButtons[self->virtualButtonCount] = GamepadAxisVirtualButton_create(
					device,
					self->inputMap->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].axisID,
					self->inputMap->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].actionID,
					self->inputMap->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].direction,
					self->inputMap->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].triggerThreshold,
					self->inputMap->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].releaseThreshold
				);
				self->virtualButtons[self->virtualButtonCount]->eventDispatcher->registerForEvent(self->virtualButtons[self->virtualButtonCount]->eventDispatcher, Atom_fromString(GAMEPAD_AXIS_VIRTUAL_BUTTON_EVENT_DOWN), virtualButtonDown, self);
				self->virtualButtons[self->virtualButtonCount]->eventDispatcher->registerForEvent(self->virtualButtons[self->virtualButtonCount]->eventDispatcher, Atom_fromString(GAMEPAD_AXIS_VIRTUAL_BUTTON_EVENT_UP), virtualButtonUp, self);
				self->virtualButtonCount++;
			}
			break;
		}
	}
	return true;
}

static bool deviceRemoved(void * sender, Atom eventID, void * eventData, void * context) {
	InputController * self = context;
	struct Gamepad_device * device = eventData;
	size_t virtualButtonIndex, deletedButtons = 0;
	
	device->eventDispatcher->unregisterForEvent(device->eventDispatcher, Atom_fromString(GAMEPAD_EVENT_BUTTON_DOWN), gamepadButtonDown, self);
	device->eventDispatcher->unregisterForEvent(device->eventDispatcher, Atom_fromString(GAMEPAD_EVENT_BUTTON_UP), gamepadButtonUp, self);
	device->eventDispatcher->unregisterForEvent(device->eventDispatcher, Atom_fromString(GAMEPAD_EVENT_AXIS_MOVED), gamepadAxisMoved, self);
	for (virtualButtonIndex = 0; virtualButtonIndex < self->virtualButtonCount; virtualButtonIndex++) {
		if (deletedButtons > 0) {
			self->virtualButtons[virtualButtonIndex] = self->virtualButtons[virtualButtonIndex + deletedButtons];
		}
		if (self->virtualButtons[virtualButtonIndex]->device == device) {
			self->virtualButtons[virtualButtonIndex]->dispose(self->virtualButtons[virtualButtonIndex]);
			deletedButtons++;
			self->virtualButtonCount--;
			virtualButtonIndex--;
		}
	}
	return true;
}

static bool axisBound(void * sender, Atom eventID, void * eventData, void * context) {
	InputController * self = context;
	struct PlayerInputMap_axisBindingEvent * event = eventData;
	size_t gamepadIndex;
	struct Gamepad_device * device;
	
	for (gamepadIndex = 0; gamepadIndex < Gamepad_numDevices(); gamepadIndex++) {
		device = Gamepad_deviceAtIndex(gamepadIndex);
		if (device->vendorID == event->vendorID && device->productID == event->productID) {
			self->virtualButtons = realloc(self->virtualButtons, sizeof(GamepadAxisVirtualButton *) * (self->virtualButtonCount + 1));
			
			self->virtualButtons[self->virtualButtonCount] = GamepadAxisVirtualButton_create(
				device,
				event->binding.axisID,
				event->binding.actionID,
				event->binding.direction,
				event->binding.triggerThreshold,
				event->binding.releaseThreshold
			);
			self->virtualButtons[self->virtualButtonCount]->eventDispatcher->registerForEvent(self->virtualButtons[self->virtualButtonCount]->eventDispatcher, Atom_fromString(GAMEPAD_AXIS_VIRTUAL_BUTTON_EVENT_DOWN), virtualButtonDown, self);
			self->virtualButtons[self->virtualButtonCount]->eventDispatcher->registerForEvent(self->virtualButtons[self->virtualButtonCount]->eventDispatcher, Atom_fromString(GAMEPAD_AXIS_VIRTUAL_BUTTON_EVENT_UP), virtualButtonUp, self);
			self->virtualButtonCount++;
		}
	}
	
	return true;
}

static bool axisUnbound(void * sender, Atom eventID, void * eventData, void * context) {
	InputController * self = context;
	struct PlayerInputMap_axisBindingEvent * event = eventData;
	size_t virtualButtonIndex;
	
	for (virtualButtonIndex = 0; virtualButtonIndex < self->virtualButtonCount; virtualButtonIndex++) {
		if (self->virtualButtons[virtualButtonIndex]->device->vendorID == event->vendorID &&
		    self->virtualButtons[virtualButtonIndex]->device->productID == event->productID &&
		    self->virtualButtons[virtualButtonIndex]->axisID == event->binding.axisID &&
		    self->virtualButtons[virtualButtonIndex]->direction == event->binding.direction &&
		    self->virtualButtons[virtualButtonIndex]->actionID == event->binding.actionID) {
			self->virtualButtons[virtualButtonIndex]->dispose(self->virtualButtons[virtualButtonIndex]);
			self->virtualButtonCount--;
			for (; virtualButtonIndex < self->virtualButtonCount; virtualButtonIndex++) {
				self->virtualButtons[virtualButtonIndex] = self->virtualButtons[virtualButtonIndex - 1];
			}
		}
	}
	
	return true;
}

void InputController_init(compat_type(InputController *) selfPtr, PlayerInputMap * inputMap) {
	InputController * self = selfPtr;
	unsigned int deviceIndex;
	
	StemObject_init(self);
	self->dispose = InputController_dispose;
	self->update = InputController_update;
	self->keyDown = InputController_keyDown;
	self->keyUp = InputController_keyUp;
	self->keyModifiersChanged = InputController_keyModifiersChanged;
	self->protected_ivar(triggerAction) = InputController_triggerAction;
	self->protected_ivar(releaseAction) = InputController_releaseAction;
	
	self->inputMap = inputMap;
	self->leftTriggered = false;
	self->rightTriggered = false;
	self->downTriggered = false;
	self->upTriggered = false;
	self->attackTriggered = false;
	self->specialTriggered = false;
	self->virtualButtonCount = 0;
	self->virtualButtons = NULL;
	self->nextDevicePoll = 1;
	
	self->inputMap->eventDispatcher->registerForEvent(self->inputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_BOUND), axisBound, self);
	self->inputMap->eventDispatcher->registerForEvent(self->inputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_UNBOUND), axisUnbound, self);
	
	Gamepad_detectDevices();
	Gamepad_processEvents();
	Gamepad_eventDispatcher()->registerForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_ATTACHED), deviceAttached, self);
	Gamepad_eventDispatcher()->registerForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_REMOVED), deviceRemoved, self);
	for (deviceIndex = 0; deviceIndex < Gamepad_numDevices(); deviceIndex++) {
		deviceAttached(NULL, GAMEPAD_EVENT_DEVICE_ATTACHED, Gamepad_deviceAtIndex(deviceIndex), self);
	}
}

void InputController_dispose(compat_type(InputController *) selfPtr) {
	InputController * self = selfPtr;
	unsigned int deviceIndex;
	
	self->inputMap->eventDispatcher->unregisterForEvent(self->inputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_BOUND), axisBound, self);
	self->inputMap->eventDispatcher->unregisterForEvent(self->inputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_UNBOUND), axisUnbound, self);
	Gamepad_eventDispatcher()->unregisterForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_ATTACHED), deviceAttached, self);
	Gamepad_eventDispatcher()->unregisterForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_REMOVED), deviceRemoved, self);
	for (deviceIndex = 0; deviceIndex < Gamepad_numDevices(); deviceIndex++) {
		deviceRemoved(NULL, GAMEPAD_EVENT_DEVICE_REMOVED, Gamepad_deviceAtIndex(deviceIndex), self);
	}
	free(self->virtualButtons);
	StemObject_dispose(self);
}

void InputController_update(compat_type(InputController *) selfPtr) {
	InputController * self = selfPtr;
	
	self->nextDevicePoll--;
	if (self->nextDevicePoll == 0) {
		Gamepad_detectDevices();
		self->nextDevicePoll = DEVICE_POLL_INTERVAL;
	}
	Gamepad_processEvents();
}

bool InputController_keyDown(compat_type(InputController *) selfPtr, unsigned int charCode, unsigned int keyCode, unsigned int modifiers) {
	InputController * self = selfPtr;
	size_t bindingIndex;
	bool handled = false;
	
	for (bindingIndex = 0; bindingIndex < self->inputMap->keyboardBindingCount; bindingIndex++) {
		if (self->inputMap->keyboardBindings[bindingIndex].keyCode == keyCode) {
			handled = self->protected_ivar(triggerAction)(self, self->inputMap->keyboardBindings[bindingIndex].actionID) || handled;
		}
	}
	return handled;
}

bool InputController_keyUp(compat_type(InputController *) selfPtr, unsigned int charCode, unsigned int keyCode, unsigned int modifiers) {
	InputController * self = selfPtr;
	size_t bindingIndex;
	bool handled = false;
	
	for (bindingIndex = 0; bindingIndex < self->inputMap->keyboardBindingCount; bindingIndex++) {
		if (self->inputMap->keyboardBindings[bindingIndex].keyCode == keyCode) {
			handled = self->protected_ivar(releaseAction)(self, self->inputMap->keyboardBindings[bindingIndex].actionID) || handled;
		}
	}
	return handled;
}

bool InputController_keyModifiersChanged(compat_type(InputController *) selfPtr, unsigned int modifiers) {
	return false;
}

bool InputController_triggerAction(compat_type(InputController *) selfPtr, Atom actionID) {
	InputController * self = selfPtr;
	
	if (actionID == Atom_fromString("left")) {
		self->leftTriggered = true;
		return true;
	}
	if (actionID == Atom_fromString("right")) {
		self->rightTriggered = true;
		return true;
	}
	if (actionID == Atom_fromString("down")) {
		self->downTriggered = true;
		return true;
	}
	if (actionID == Atom_fromString("up")) {
		self->upTriggered = true;
		return true;
	}
	if (actionID == Atom_fromString("attack")) {
		self->attackTriggered = true;
		return true;
	}
	if (actionID == Atom_fromString("special")) {
		self->specialTriggered = true;
		return true;
	}
	return false;
}

bool InputController_releaseAction(compat_type(InputController *) selfPtr, Atom actionID) {
	InputController * self = selfPtr;
	
	if (actionID == Atom_fromString("left")) {
		self->leftTriggered = false;
		return true;
	}
	if (actionID == Atom_fromString("right")) {
		self->rightTriggered = false;
		return true;
	}
	if (actionID == Atom_fromString("down")) {
		self->downTriggered = false;
		return true;
	}
	if (actionID == Atom_fromString("up")) {
		self->upTriggered = false;
		return true;
	}
	if (actionID == Atom_fromString("attack")) {
		self->attackTriggered = false;
		return true;
	}
	if (actionID == Atom_fromString("special")) {
		self->specialTriggered = false;
		return true;
	}
	return false;
}
