/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/MovingEntityView.h"
#include <math.h>
#include <stddef.h>
#include <stdlib.h>

struct vertex_p2f {
	GLfloat position[2];
};

MovingEntityView * MovingEntityView_create(compat_type(MovingEntityModel *) modelPtr) {
	stemobject_create_implementation(MovingEntityView, init, modelPtr)
}

#define TESSELATIONS 32
#define OUTLINE_WIDTH 0.0625f

void MovingEntityView_init(compat_type(MovingEntityView *) selfPtr, compat_type(MovingEntityModel *) modelPtr) {
	MovingEntityView * self = selfPtr;
	struct vertex_p2f vertices[TESSELATIONS * 2 + 3];
	GLubyte indexes[(TESSELATIONS - 2) * 3 + TESSELATIONS * 6 + 3];
	unsigned int vertexIndex;
	MovingEntityModel * model = modelPtr;
	
	GameObjectView_init(self, modelPtr);
	self->dispose = MovingEntityView_dispose;
	self->draw = MovingEntityView_draw;
	
	glGenBuffers(1, &self->vboID);
	glGenBuffers(1, &self->eboID);
	for (vertexIndex = 0; vertexIndex < TESSELATIONS; vertexIndex++) {
		vertices[vertexIndex].position[0] = cos(M_PI * 2 * (vertexIndex + 0.5f) / TESSELATIONS) * (model->radius - OUTLINE_WIDTH);
		vertices[vertexIndex].position[1] = sin(M_PI * 2 * (vertexIndex + 0.5f) / TESSELATIONS) * (model->radius - OUTLINE_WIDTH);
		if (vertexIndex >= 2) {
			indexes[(vertexIndex - 2) * 3 + 0] = 0;
			indexes[(vertexIndex - 2) * 3 + 1] = vertexIndex - 1;
			indexes[(vertexIndex - 2) * 3 + 2] = vertexIndex;
		}
	}
	for (vertexIndex = 0; vertexIndex < TESSELATIONS; vertexIndex++) {
		vertices[TESSELATIONS + vertexIndex].position[0] = cos(M_PI * 2 * (vertexIndex + 0.5f) / TESSELATIONS) * model->radius;
		vertices[TESSELATIONS + vertexIndex].position[1] = sin(M_PI * 2 * (vertexIndex + 0.5f) / TESSELATIONS) * model->radius;
		indexes[(TESSELATIONS - 2) * 3 + vertexIndex * 6 + 0] = vertexIndex;
		indexes[(TESSELATIONS - 2) * 3 + vertexIndex * 6 + 1] = TESSELATIONS + vertexIndex;
		indexes[(TESSELATIONS - 2) * 3 + vertexIndex * 6 + 2] = (vertexIndex + 1) % TESSELATIONS;
		indexes[(TESSELATIONS - 2) * 3 + vertexIndex * 6 + 3] = (vertexIndex + 1) % TESSELATIONS;
		indexes[(TESSELATIONS - 2) * 3 + vertexIndex * 6 + 4] = TESSELATIONS + (vertexIndex + 1) % TESSELATIONS;
		indexes[(TESSELATIONS - 2) * 3 + vertexIndex * 6 + 5] = TESSELATIONS + vertexIndex;
	}
	vertices[TESSELATIONS * 2 + 0].position[0] = 0.0f;
	vertices[TESSELATIONS * 2 + 0].position[1] = model->radius + 0.25f;
	vertices[TESSELATIONS * 2 + 1].position[0] = -0.125f;
	vertices[TESSELATIONS * 2 + 1].position[1] = model->radius + 0.125f;
	vertices[TESSELATIONS * 2 + 2].position[0] = 0.125f;
	vertices[TESSELATIONS * 2 + 2].position[1] = model->radius + 0.125f;
	indexes[(TESSELATIONS - 2) * 3 + TESSELATIONS * 6 + 0] = TESSELATIONS * 2 + 0;
	indexes[(TESSELATIONS - 2) * 3 + TESSELATIONS * 6 + 1] = TESSELATIONS * 2 + 1;
	indexes[(TESSELATIONS - 2) * 3 + TESSELATIONS * 6 + 2] = TESSELATIONS * 2 + 2;
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f) * (TESSELATIONS * 2 + 3), vertices, GL_STATIC_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLubyte) * ((TESSELATIONS - 2) * 3 + TESSELATIONS * 6 + 3), indexes, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void MovingEntityView_dispose(compat_type(MovingEntityView *) selfPtr) {
	MovingEntityView * self = selfPtr;
	
	glDeleteBuffers(1, &self->vboID);
	glDeleteBuffers(1, &self->eboID);
	GameObjectView_dispose(self);
}

void MovingEntityView_draw(compat_type(MovingEntityView *) selfPtr, Matrix modelviewMatrix, double deltaTime) {
	MovingEntityView * self = selfPtr;
	MovingEntityModel * model = (MovingEntityModel *) self->model;
	
	Matrix_translate(&modelviewMatrix, model->position.x, model->position.y, 0.0f);
	Matrix_rotate(&modelviewMatrix, Vector3_init(0.0f, 0.0f, -1.0f), directionToRadians(model->facing));
	
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	
	glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f), (void *) offsetof(struct vertex_p2f, position));
	glEnableClientState(GL_VERTEX_ARRAY);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	glColor4ub(self->innerColor >> 24 & 0xFF, self->innerColor >> 16 & 0xFF, self->innerColor >> 8 & 0xFF, self->innerColor & 0xFF);
	glDrawElements(GL_TRIANGLES, (TESSELATIONS - 2) * 3, GL_UNSIGNED_BYTE, 0);
	glColor4ub(self->outerColor >> 24 & 0xFF, self->outerColor >> 16 & 0xFF, self->outerColor >> 8 & 0xFF, self->outerColor & 0xFF);
	glDrawElements(GL_TRIANGLES, TESSELATIONS * 6 + 3, GL_UNSIGNED_BYTE, (GLvoid *) ((TESSELATIONS - 2) * 3));
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
