/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/PlaybackPlayerInputController.h"
#include "jsonserialization/JSONDeserializationContext.h"
#include <stdio.h>

PlaybackPlayerInputController * PlaybackPlayerInputController_create(PlayerModel * playerModel, PlayerInputMap * inputMap, const char * fileName) {
	stemobject_create_implementation(PlaybackPlayerInputController, init, playerModel, inputMap, fileName)
}

void PlaybackPlayerInputController_init(compat_type(PlaybackPlayerInputController *) selfPtr, PlayerModel * playerModel, PlayerInputMap * inputMap, const char * fileName) {
	PlaybackPlayerInputController * self = selfPtr;
	JSONDeserializationContext * context;
	size_t nodeIndex;
	uint16_t formatVersion;
	size_t nodeCount;
	struct RecordedPlayerInputNode * nodes;
	
	PlayerInputController_init(self, playerModel, inputMap);
	self->dispose = PlaybackPlayerInputController_dispose;
	self->protected_ivar(triggerAction) = PlaybackPlayerInputController_triggerAction;
	self->protected_ivar(releaseAction) = PlaybackPlayerInputController_releaseAction;
	self->update = PlaybackPlayerInputController_update;
	
	self->frameIndex = 0;
	self->nodeIndex = 0;
	self->nodeCount = 0;
	self->nodes = NULL;
	
	context = JSONDeserializationContext_createWithFile(fileName);
	if (context->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Couldn't load \"%s\" as a json file; if you've specified it as a relative path, an absolute path may work\n", fileName);
		return;
	}
	context->beginStructure(context, "input_session");
	formatVersion = context->readUInt16(context, "format_version");
	if (formatVersion != PLAYER_INPUT_SESSION_VERSION) {
		fprintf(stderr, "Error: Input session file \"%s\" is of version %d; requires %d\n", fileName, formatVersion, PLAYER_INPUT_SESSION_VERSION);
		context->dispose(context);
		return;
	}
	nodeCount = context->beginArray(context, "events");
	nodes = malloc(sizeof(struct RecordedPlayerInputNode) * nodeCount);
	for (nodeIndex = 0; nodeIndex < nodeCount; nodeIndex++) {
		context->beginStructure(context, NULL);
		nodes[nodeIndex].frameIndex = context->readUInt32(context, "frame");
		nodes[nodeIndex].actionID = Atom_fromString(context->readString(context, "action"));
		nodes[nodeIndex].triggered = context->readBoolean(context, "triggered");
		context->endStructure(context);
	}
	context->endArray(context);
	context->endStructure(context);
	
	if (context->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Error: Failed to load \"%s\" as input session file (error %d)\n", fileName, context->status);
		free(nodes);
	} else {
		self->nodeCount = nodeCount;
		self->nodes = nodes;
	}
	context->dispose(context);
}

void PlaybackPlayerInputController_dispose(compat_type(PlaybackPlayerInputController *) selfPtr) {
	PlaybackPlayerInputController * self = selfPtr;
	
	free(self->nodes);
	PlayerInputController_dispose(self);
}

bool PlaybackPlayerInputController_triggerAction(compat_type(PlaybackPlayerInputController *) selfPtr, Atom actionID) {
	return false;
}

bool PlaybackPlayerInputController_releaseAction(compat_type(PlaybackPlayerInputController *) selfPtr, Atom actionID) {
	return false;
}

void PlaybackPlayerInputController_update(compat_type(PlaybackPlayerInputController *) selfPtr) {
	PlaybackPlayerInputController * self = selfPtr;
	
	while (self->nodeIndex < self->nodeCount && self->nodes[self->nodeIndex].frameIndex == self->frameIndex) {
		if (self->nodes[self->nodeIndex].triggered) {
			PlayerInputController_triggerAction(self, self->nodes[self->nodeIndex].actionID);
		} else {
			PlayerInputController_releaseAction(self, self->nodes[self->nodeIndex].actionID);
		}
		self->nodeIndex++;
		if (self->nodeIndex >= self->nodeCount) {
			printf("Playback complete\n");
		}
	}
	self->frameIndex++;
	InputController_update(self);
}
