/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/PlayerDuelistModel.h"
#include "convergence/ProjectileModel.h"
#include "convergence/SharedDefinitions.h"
#include "convergence/SwordSwingModel.h"

#define BLOCK_BUFF_DURATION 180

static const int BLOCK_ABSORBS[3] = {3, 4, 5};
static const float BLOCK_KNOCKBACKS[3] = {0.75f, 0.625f, 0.5f};

PlayerDuelistModel * PlayerDuelistModel_create(DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	stemobject_create_implementation(PlayerDuelistModel, init, stateModel, position, facing)
}

static void sharedInit(PlayerDuelistModel * self) {
	self->dispose = PlayerDuelistModel_dispose;
	self->copy = PlayerDuelistModel_copy;
	self->update = PlayerDuelistModel_update;
	self->atRest = PlayerDuelistModel_atRest;
	self->startAttack = PlayerDuelistModel_startAttack;
	self->startSpecialAction = PlayerDuelistModel_startSpecialAction;
	self->stopSpecialAction = PlayerDuelistModel_stopSpecialAction;
	self->protected_ivar(knockedBack) = PlayerDuelistModel_knockedBack;
	self->protected_ivar(takeDamage) = PlayerDuelistModel_takeDamage;
	
	self->playerClass = PLAYER_CLASS_DUELIST;
	self->acceleration = 1.0f / 256.0f;
	self->deceleration = 2.0f / 256.0f;
	self->velocityMax = 14.0f / 256.0f;
}

void PlayerDuelistModel_init(compat_type(PlayerDuelistModel *) selfPtr, DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	PlayerDuelistModel * self = selfPtr;
	
	PlayerModel_init(self, stateModel, position, facing);
	sharedInit(self);
	self->attackCooldown = 0;
	self->blocking = false;
	self->blockBuffRemaining = 0;
}

void PlayerDuelistModel_initCopy(compat_type(PlayerDuelistModel *) selfPtr, compat_type(PlayerDuelistModel *) originalPtr) {
	PlayerDuelistModel * self = selfPtr, * original = originalPtr;
	
	PlayerModel_initCopy(self, original);
	sharedInit(self);
	self->attackCooldown = original->attackCooldown;
	self->swordSwingFacing = original->swordSwingFacing;
	self->blocking = original->blocking;
	self->blockFacing = original->blockFacing;
	self->blockBuffRemaining = original->blockBuffRemaining;
	self->velocityMax = original->velocityMax;
}

void PlayerDuelistModel_dispose(compat_type(PlayerDuelistModel *) selfPtr) {
	PlayerDuelistModel * self = selfPtr;
	
	PlayerModel_dispose(self);
}

compat_type(PlayerDuelistModel *) PlayerDuelistModel_copy(compat_type(PlayerDuelistModel *) selfPtr) {
	stemobject_copy_implementation(PlayerDuelistModel, initCopy)
}

void PlayerDuelistModel_update(compat_type(PlayerDuelistModel *) selfPtr) {
	PlayerDuelistModel * self = selfPtr;
	
	PlayerModel_update(self);
	
	if (self->attackCooldown > 0) {
		self->attackCooldown--;
		self->facing = self->swordSwingFacing;
	}
	if (self->blockBuffRemaining > 0) {
		self->blockBuffRemaining--;
	}
	if (self->blocking) {
		self->facing = self->blockFacing;
	}
}

bool PlayerDuelistModel_atRest(compat_type(PlayerDuelistModel *) selfPtr) {
	PlayerDuelistModel * self = selfPtr;
	return self->attackCooldown == 0 &&
	       PlayerModel_atRest(self);
}

void PlayerDuelistModel_startAttack(compat_type(PlayerDuelistModel *) selfPtr) {
	PlayerDuelistModel * self = selfPtr;
	
	if (self->attackCooldown == 0) {
		self->swordSwingFacing = self->facing;
		switch (self->rank) {
			case 0:
			default:
				self->attackCooldown = 32;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, false, self->blockBuffRemaining > 0 ? 3 : 1, 0.125f, 0.5f, 2.0f, 0.625f, 32));
				break;
			case 1:
				self->attackCooldown = 28;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, false, self->blockBuffRemaining > 0 ? 5 : 2, 0.125f, 0.5f, 2.125f, 0.625f, 28));
				break;
			case 2:
				self->attackCooldown = 26;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, false, self->blockBuffRemaining > 0 ? 8 : 3, 0.125f, 0.5f, 2.25f, 0.625f, 26));
				break;
		}
		self->blockBuffRemaining = 0;
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_ATTACK_BIG), NULL);
	}
}

void PlayerDuelistModel_startSpecialAction(compat_type(PlayerDuelistModel *) selfPtr) {
	PlayerDuelistModel * self = selfPtr;
	
	self->blocking = true;
	self->blockFacing = self->facing;
	self->velocityMax = 9.0f / 256.0f;
}

void PlayerDuelistModel_stopSpecialAction(compat_type(PlayerDuelistModel *) selfPtr) {
	PlayerDuelistModel * self = selfPtr;
	
	self->blocking = false;
	self->velocityMax = 14.0f / 256.0f;
}

void PlayerDuelistModel_knockedBack(compat_type(PlayerDuelistModel *) selfPtr, Vector2 knockback) {
	PlayerDuelistModel * self = selfPtr;
	
	PlayerModel_knockedBack(selfPtr, Vector2_multiplyScalar(knockback, BLOCK_KNOCKBACKS[self->rank]));
}

void PlayerDuelistModel_takeDamage(compat_type(PlayerDuelistModel *) selfPtr, int amount, Direction fromDirection) {
	PlayerDuelistModel * self = selfPtr;
	
	if (self->blocking && self->attackCooldown == 0 && Vector2_dot(directionToVector(self->facing), directionToVector(fromDirection)) > -0.00001f) {
		PlayerModel_takeDamage(self, max(0, amount - BLOCK_ABSORBS[self->rank]), fromDirection);
		self->hitCooldown = 10;
		self->blockBuffRemaining = BLOCK_BUFF_DURATION;
	} else {
		PlayerModel_takeDamage(self, amount, fromDirection);
		self->blockBuffRemaining = 0;
	}
}
