/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/PlayerRangerModel.h"
#include "convergence/ProjectileModel.h"
#include "convergence/SwordSwingModel.h"

PlayerRangerModel * PlayerRangerModel_create(DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	stemobject_create_implementation(PlayerRangerModel, init, stateModel, position, facing)
}

static void sharedInit(PlayerRangerModel * self) {
	self->dispose = PlayerRangerModel_dispose;
	self->copy = PlayerRangerModel_copy;
	self->update = PlayerRangerModel_update;
	self->atRest = PlayerRangerModel_atRest;
	self->startAttack = PlayerRangerModel_startAttack;
	self->startSpecialAction = PlayerRangerModel_startSpecialAction;
	
	self->playerClass = PLAYER_CLASS_RANGER;
	self->acceleration = 1.0f / 256.0f;
	self->deceleration = 2.0f / 256.0f;
	self->velocityMax = 15.0f / 256.0f;
}

void PlayerRangerModel_init(compat_type(PlayerRangerModel *) selfPtr, DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	PlayerRangerModel * self = selfPtr;
	
	PlayerModel_init(self, stateModel, position, facing);
	sharedInit(self);
	self->attackCooldown = 0;
	self->facingLocked = false;
}

void PlayerRangerModel_initCopy(compat_type(PlayerRangerModel *) selfPtr, compat_type(PlayerRangerModel *) originalPtr) {
	PlayerRangerModel * self = selfPtr, * original = originalPtr;
	
	PlayerModel_initCopy(self, original);
	sharedInit(self);
	self->attackCooldown = original->attackCooldown;
	self->facingLocked = original->facingLocked;
	self->lockedFacing = original->lockedFacing;
}

void PlayerRangerModel_dispose(compat_type(PlayerRangerModel *) selfPtr) {
	PlayerRangerModel * self = selfPtr;
	
	PlayerModel_dispose(self);
}

compat_type(PlayerRangerModel *) PlayerRangerModel_copy(compat_type(PlayerRangerModel *) selfPtr) {
	stemobject_copy_implementation(PlayerRangerModel, initCopy)
}

void PlayerRangerModel_update(compat_type(PlayerRangerModel *) selfPtr) {
	PlayerRangerModel * self = selfPtr;
	
	PlayerModel_update(self);
	
	if (self->attackCooldown > 0) {
		self->attackCooldown--;
	}
	if (self->facingLocked) {
		self->facing = self->lockedFacing;
	}
}

bool PlayerRangerModel_atRest(compat_type(PlayerRangerModel *) selfPtr) {
	PlayerRangerModel * self = selfPtr;
	return self->attackCooldown == 0 &&
	       PlayerModel_atRest(self);
}

static Vector2 Vector2_rotate(Vector2 vector, float angle) {
	Vector2 result;
	
	result.x = vector.x * cos(angle) + vector.y * -sin(angle);
	result.y = vector.x * sin(angle) + vector.y * cos(angle);
	return result;
}

void PlayerRangerModel_startAttack(compat_type(PlayerRangerModel *) selfPtr) {
	PlayerRangerModel * self = selfPtr;
	
	if (self->attackCooldown == 0 && self->hitCooldown == 0) {
		ProjectileModel * projectile;
		
		switch (self->rank) {
			case 0:
			default:
				self->attackCooldown = 12;
				projectile = ProjectileModel_create(self->stateModel, self->position, directionToVector(self->facing), 0.125f, 7.0f, false, 1, 0.03125f);
				break;
			case 1:
				self->attackCooldown = 10;
				projectile = ProjectileModel_create(self->stateModel, self->position, directionToVector(self->facing), 0.140625f, 8.0f, false, 2, 0.046875f);
				break;
			case 2:
				self->attackCooldown = 8;
				projectile = ProjectileModel_create(self->stateModel, self->position, Vector2_rotate(directionToVector(self->facing), -M_PI * 0.03125f), 0.15625f, 8.5f, false, 3, 0.046875f);
				projectile->position.x += sin(directionToRadians(self->facing) - M_PI * 0.03125f) * (self->radius - projectile->radius);
				projectile->position.y += cos(directionToRadians(self->facing) - M_PI * 0.03125f) * (self->radius - projectile->radius);
				self->stateModel->addObject(self->stateModel, projectile);
				projectile = ProjectileModel_create(self->stateModel, self->position, Vector2_rotate(directionToVector(self->facing), M_PI * 0.03125f), 0.15625f, 8.5f, false, 3, 0.046875f);
				projectile->position.x += sin(directionToRadians(self->facing) + M_PI * 0.03125f) * (self->radius - projectile->radius);
				projectile->position.y += cos(directionToRadians(self->facing) + M_PI * 0.03125f) * (self->radius - projectile->radius);
				self->stateModel->addObject(self->stateModel, projectile);
				projectile = ProjectileModel_create(self->stateModel, self->position, directionToVector(self->facing), 0.15625f, 8.5f, false, 3, 0.046875f);
				break;
		}
		projectile->position.x += sin(directionToRadians(self->facing)) * (self->radius - projectile->radius);
		projectile->position.y += cos(directionToRadians(self->facing)) * (self->radius - projectile->radius);
		self->stateModel->addObject(self->stateModel, projectile);
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_SHOOT), NULL);
	}
}

void PlayerRangerModel_startSpecialAction(compat_type(PlayerModel *) selfPtr) {
	PlayerRangerModel * self = selfPtr;
	
	self->facingLocked = !self->facingLocked;
	if (self->facingLocked) {
		self->lockedFacing = self->facing;
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANGER_LOCK), NULL);
	}
}
