/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/ProjectileModel.h"
#include <math.h>
#include <stdlib.h>

ProjectileModel * ProjectileModel_create(DungeonRoomStateModel * stateModel, Vector2 position, Vector2 direction, float speed, float travelDistance, bool hostile, int damage, float knockback) {
	stemobject_create_implementation(ProjectileModel, init, stateModel, position, direction, speed, travelDistance, hostile, damage, knockback)
}

static void sharedInit(ProjectileModel * self) {
	self->dispose = ProjectileModel_dispose;
	self->copy = ProjectileModel_copy;
	self->update = ProjectileModel_update;
	self->atRest = ProjectileModel_atRest;
	self->roughBounds = ProjectileModel_roughBounds;
	self->collisionGeometry = ProjectileModel_collisionGeometry;
	self->wouldCollide = ProjectileModel_wouldCollide;
	self->collision = ProjectileModel_collision;
	
	self->type = GAME_OBJECT_TYPE_PROJECTILE;
	self->radius = 0.125f;
	self->distanceTraveled = 0.0f;
}

void ProjectileModel_init(compat_type(ProjectileModel *) selfPtr, DungeonRoomStateModel * stateModel, Vector2 position, Vector2 direction, float speed, float travelDistance, bool hostile, int damage, float knockback) {
	ProjectileModel * self = selfPtr;
	
	GameObjectModel_init(self, stateModel);
	sharedInit(self);
	
	self->lastPosition = self->position = position;
	self->direction = direction;
	self->speed = speed;
	self->travelDistance = travelDistance;
	self->hostile = hostile;
	self->damage = damage;
	self->knockback = knockback;
}

void ProjectileModel_initCopy(compat_type(ProjectileModel *) selfPtr, compat_type(ProjectileModel *) originalPtr) {
	ProjectileModel * self = selfPtr, * original = originalPtr;
	
	GameObjectModel_initCopy(self, original);
	sharedInit(self);
	
	self->lastPosition = original->lastPosition;
	self->position = original->position;
	self->direction = original->direction;
	self->radius = original->radius;
	self->travelDistance = original->travelDistance;
	self->distanceTraveled = original->distanceTraveled;
	self->speed = original->speed;
	self->hostile = original->hostile;
	self->damage = original->damage;
	self->knockback = original->knockback;
}

void ProjectileModel_dispose(compat_type(ProjectileModel *) selfPtr) {
	ProjectileModel * self = selfPtr;
	
	GameObjectModel_dispose(self);
}

compat_type(ProjectileModel *) ProjectileModel_copy(compat_type(ProjectileModel *) selfPtr) {
	stemobject_copy_implementation(ProjectileModel, initCopy)
}

void ProjectileModel_update(compat_type(ProjectileModel *) selfPtr) {
	ProjectileModel * self = selfPtr;
	
	self->lastPosition = self->position;
	self->position.x += self->speed * self->direction.x;
	self->position.y += self->speed * self->direction.y;
	self->distanceTraveled += self->speed;
	if (self->distanceTraveled >= self->travelDistance) {
		self->markedForRemoval = true;
	}
}

bool ProjectileModel_atRest(compat_type(ProjectileModel *) selfPtr) {
	return false;
}

struct rect4f ProjectileModel_roughBounds(compat_type(ProjectileModel *) selfPtr) {
	ProjectileModel * self = selfPtr;
	struct rect4f bounds;
	
	bounds.left = min(self->lastPosition.x, self->position.x) - self->radius;
	bounds.right = max(self->lastPosition.x, self->position.x) + self->radius;
	bounds.bottom = min(self->lastPosition.y, self->position.y) - self->radius;
	bounds.top = max(self->lastPosition.y, self->position.y) + self->radius;
	
	return bounds;
}

struct collisionGeometry ProjectileModel_collisionGeometry(compat_type(ProjectileModel *) selfPtr) {
	ProjectileModel * self = selfPtr;
	struct collisionGeometry geometry;
	
	geometry.type = COLLISION_GEOMETRY_TYPE_SWEPT_AABB;
	geometry.u.sweptAABB.current.x = self->lastPosition.x - self->radius;
	geometry.u.sweptAABB.current.y = self->lastPosition.y - self->radius;
	geometry.u.sweptAABB.future.x = self->position.x - self->radius;
	geometry.u.sweptAABB.future.y = self->position.y - self->radius;
	geometry.u.sweptAABB.width = geometry.u.sweptAABB.height = self->radius * 2.0f;
	return geometry;
}

bool ProjectileModel_wouldCollide(compat_type(ProjectileModel *) selfPtr, compat_type(GameObjectModel *) objectPtr) {
	ProjectileModel * self = selfPtr;
	
	if (self->markedForRemoval) {
		return false;
	}
	switch (((GameObjectModel *) objectPtr)->type) {
		case GAME_OBJECT_TYPE_ENEMY:
			return !self->hostile;
		case GAME_OBJECT_TYPE_PLAYER:
			return self->hostile;
		case GAME_OBJECT_TYPE_DOOR:
			return true;
		default:
			return false;
	}
}

void ProjectileModel_collision(compat_type(ProjectileModel *) selfPtr, compat_type(GameObjectModel *) objectPtr, const struct collision collision, float time) {
	ProjectileModel * self = selfPtr;
	GameObjectModel * object = objectPtr;
	
	self->position.x = self->lastPosition.x + (self->position.x - self->lastPosition.x) * collision.time;
	self->position.y = self->lastPosition.y + (self->position.y - self->lastPosition.y) * collision.time;
	self->markedForRemoval = true;
	if (object == NULL || object->type == GAME_OBJECT_TYPE_DOOR) {
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PROJECTILE_EVENT_SHOT_STOPPED_WALL), NULL);
	}
}

void ProjectileModel_interpolate(compat_type(ProjectileModel *) selfPtr, float time) {
	ProjectileModel * self = selfPtr;
	
	self->lastPosition.x += (self->position.x - self->lastPosition.x) * time;
	self->lastPosition.y += (self->position.y - self->lastPosition.y) * time;
}
