/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/RainbowTrailEffect.h"
#include "convergence/SharedDefinitions.h"
#include "shell/Shell.h"
#include <stddef.h>
#include <stdlib.h>

struct vertex_p2f_c4f_t2f {
	GLfloat position[2];
	GLfloat color[4];
	GLfloat texCoords[2];
};

RainbowTrailEffect * RainbowTrailEffect_create(compat_type(PlayerSpeedsterModel *) player, ResourceManager * resourceManager) {
	stemobject_create_implementation(RainbowTrailEffect, init, player, resourceManager)
}

static bool playerDied(void * sender, Atom eventID, void * eventData, void * context) {
	RainbowTrailEffect * self = context;
	
	self->player->eventDispatcher->registerForEvent(self->player->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDied, self);
	self->player = NULL;
	return true;
}

void RainbowTrailEffect_init(compat_type(RainbowTrailEffect *) selfPtr, compat_type(PlayerSpeedsterModel *) player, ResourceManager * resourceManager) {
	RainbowTrailEffect * self = selfPtr;
	
	GameObjectView_init(self, NULL);
	self->dispose = RainbowTrailEffect_dispose;
	self->draw = RainbowTrailEffect_draw;
	
	self->resourceManager = resourceManager;
	self->player = player;
	self->player->eventDispatcher->registerForEvent(self->player->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDied, self);
	self->texture = self->resourceManager->referenceResource(self->resourceManager, "texture", "textures/rainbow.json");
	
	self->sortOrder = -1;
	self->nodeCount = 1;
	self->nodes[0].time = 0.0;
	self->lastPlayerPosition = self->player->position;
	self->currentTime = 0.0;
	
	glGenBuffers(1, &self->vboID);
	glGenBuffers(1, &self->eboID);
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_c4f_t2f) * 2 * RAINBOW_TRAIL_EFFECT_NODE_MAX, NULL, GL_STREAM_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * 6 * (RAINBOW_TRAIL_EFFECT_NODE_MAX - 1), NULL, GL_STREAM_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void RainbowTrailEffect_dispose(compat_type(RainbowTrailEffect *) selfPtr) {
	RainbowTrailEffect * self = selfPtr;
	
	if (self->player != NULL) {
		self->player->eventDispatcher->unregisterForEvent(self->player->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDied, self);
	}
	self->resourceManager->releaseResource(self->resourceManager, "texture", "textures/rainbow.json");
	glDeleteBuffers(1, &self->vboID);
	glDeleteBuffers(1, &self->eboID);
	GameObjectView_dispose(self);
}

#define RADIUS 0.4375f
#define TRAIL_DURATION 0.1875f
#define FADE_DURATION 0.125f
#define NODE_FADE_DURATION 0.1875f

void RainbowTrailEffect_draw(compat_type(RainbowTrailEffect *) selfPtr, Matrix modelviewMatrix, double deltaTime) {
	RainbowTrailEffect * self = selfPtr;
	float maxAlpha;
	struct vertex_p2f_c4f_t2f * vertices;
	struct vertex_p2f_c4f_t2f vertex;
	GLushort * indexes;
	int nodeIndex;
	
	self->currentTime += deltaTime;
	if (self->currentTime >= TRAIL_DURATION + FADE_DURATION) {
		self->markedForRemoval = true;
	}
	
	if (self->currentTime >= TRAIL_DURATION) {
		maxAlpha = 1.0f - (self->currentTime - TRAIL_DURATION) / FADE_DURATION;
	} else {
		maxAlpha = 1.0f;
	}
	
	if (self->player != NULL && self->nodeCount < RAINBOW_TRAIL_EFFECT_NODE_MAX && self->currentTime < TRAIL_DURATION + FADE_DURATION && (self->player->position.x != self->lastPlayerPosition.x || self->player->position.y != self->lastPlayerPosition.y)) {
		Vector2 direction;
		
		direction = Vector2_subtract(self->player->position, self->lastPlayerPosition);
		Vector2_normalize(&direction);
		self->nodes[self->nodeCount].time = self->currentTime;
		self->nodes[self->nodeCount].leftVertex.x = self->player->position.x + direction.y * RADIUS;
		self->nodes[self->nodeCount].leftVertex.y = self->player->position.y - direction.x * RADIUS;
		self->nodes[self->nodeCount].rightVertex.x = self->player->position.x - direction.y * RADIUS;
		self->nodes[self->nodeCount].rightVertex.y = self->player->position.y + direction.x * RADIUS;
		if (self->nodeCount == 1) {
			self->nodes[0].leftVertex.x = self->lastPlayerPosition.x + direction.y * RADIUS;
			self->nodes[0].leftVertex.y = self->lastPlayerPosition.y - direction.x * RADIUS;
			self->nodes[0].rightVertex.x = self->lastPlayerPosition.x - direction.y * RADIUS;
			self->nodes[0].rightVertex.y = self->lastPlayerPosition.y + direction.x * RADIUS;
		}
		self->nodeCount++;
		self->lastPlayerPosition = self->player->position;
	}
	
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	
	vertices = glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
	indexes = glMapBuffer(GL_ELEMENT_ARRAY_BUFFER, GL_WRITE_ONLY);
	vertex.color[0] = vertex.color[1] = vertex.color[2] = 1.0f;
	vertex.texCoords[0] = vertex.texCoords[1] = 0.0f;
	vertex.position[0] = self->nodes[0].leftVertex.x;
	vertex.position[1] = self->nodes[0].leftVertex.y;
	vertex.color[3] = max(0.0f, 1.0f - (self->currentTime - self->nodes[0].time) / NODE_FADE_DURATION) * maxAlpha;
	vertices[0] = vertex;
	vertex.position[0] = self->nodes[0].rightVertex.x;
	vertex.position[1] = self->nodes[0].rightVertex.y;
	vertex.texCoords[0] = 1.0f;
	vertices[1] = vertex;
	for (nodeIndex = 1; nodeIndex < self->nodeCount; nodeIndex++) {
		vertex.position[0] = self->nodes[nodeIndex].leftVertex.x;
		vertex.position[1] = self->nodes[nodeIndex].leftVertex.y;
		vertex.texCoords[0] = 0.0f;
		vertex.color[3] = max(0.0f, 1.0f - (self->currentTime - self->nodes[nodeIndex].time) / NODE_FADE_DURATION) * maxAlpha;
		vertices[nodeIndex * 2 + 0] = vertex;
		vertex.position[0] = self->nodes[nodeIndex].rightVertex.x;
		vertex.position[1] = self->nodes[nodeIndex].rightVertex.y;
		vertex.texCoords[0] = 1.0f;
		vertices[nodeIndex * 2 + 1] = vertex;
		
		indexes[(nodeIndex - 1) * 6 + 0] = nodeIndex * 2 - 2;
		indexes[(nodeIndex - 1) * 6 + 1] = nodeIndex * 2 - 1;
		indexes[(nodeIndex - 1) * 6 + 2] = nodeIndex * 2 + 1;
		indexes[(nodeIndex - 1) * 6 + 3] = nodeIndex * 2 + 1;
		indexes[(nodeIndex - 1) * 6 + 4] = nodeIndex * 2 + 0;
		indexes[(nodeIndex - 1) * 6 + 5] = nodeIndex * 2 - 2;
	}
	glUnmapBuffer(GL_ARRAY_BUFFER);
	glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);
	
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_c4f_t2f), (void *) offsetof(struct vertex_p2f_c4f_t2f, position));
	glColorPointer(4, GL_FLOAT, sizeof(struct vertex_p2f_c4f_t2f), (void *) offsetof(struct vertex_p2f_c4f_t2f, color));
	glTexCoordPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_c4f_t2f), (void *) offsetof(struct vertex_p2f_c4f_t2f, texCoords));
	
	self->texture->activate(self->texture);
	glDrawElements(GL_TRIANGLES, 6 * (self->nodeCount - 1), GL_UNSIGNED_SHORT, 0);
	self->texture->deactivate(self->texture);
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
