/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/SparkleEffect.h"
#include "shell/Shell.h"
#include <stddef.h>
#include <stdlib.h>

struct vertex_p2f_c4ub {
	GLfloat position[2];
	GLubyte color[4];
};

#define DURATION 1.0f
#define STAR_OFFSET 1.25f
#define STAR_RADIUS_0 0.1875f
#define STAR_RADIUS_1 0.75f
#define STAR_RADIUS_2 1.0f
#define RATIO 0.875f

SparkleEffect * SparkleEffect_create(Vector2 position, float radius, compat_type(PlayerModel *) owner) {
	stemobject_create_implementation(SparkleEffect, init, position, radius, owner)
}

static bool playerDeath(void * sender, Atom eventID, void * eventData, void * context) {
	SparkleEffect * self = context;
	
	self->owner->eventDispatcher->unregisterForEvent(self->owner->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
	self->owner = NULL;
	return true;
}

void SparkleEffect_init(compat_type(SparkleEffect *) selfPtr, Vector2 position, float radius, compat_type(PlayerModel *) owner) {
	SparkleEffect * self = selfPtr;
	GLushort indexes[210];
	int starIndex, pointIndex;
	
	GameObjectView_init(self, NULL);
	self->dispose = SparkleEffect_dispose;
	self->draw = SparkleEffect_draw;
	
	self->sortOrder = 1;
	self->position = position;
	self->radius = radius;
	self->owner = owner;
	if (self->owner != NULL) {
		self->owner->eventDispatcher->registerForEvent(self->owner->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
	}
	self->currentTime = 0.0;
	
	for (starIndex = 0; starIndex < 7; starIndex++) {
		for (pointIndex = 0; pointIndex < 6; pointIndex++) {
			indexes[starIndex * 30 + pointIndex * 3 + 0] = starIndex * 12 + (pointIndex * 2 + 0) % 12;
			indexes[starIndex * 30 + pointIndex * 3 + 1] = starIndex * 12 + (pointIndex * 2 + 1) % 12;
			indexes[starIndex * 30 + pointIndex * 3 + 2] = starIndex * 12 + (pointIndex * 2 + 2) % 12;
		}
		indexes[starIndex * 30 + 18] = starIndex * 12 + 0;
		indexes[starIndex * 30 + 19] = starIndex * 12 + 2;
		indexes[starIndex * 30 + 20] = starIndex * 12 + 4;
		indexes[starIndex * 30 + 21] = starIndex * 12 + 0;
		indexes[starIndex * 30 + 22] = starIndex * 12 + 4;
		indexes[starIndex * 30 + 23] = starIndex * 12 + 6;
		indexes[starIndex * 30 + 24] = starIndex * 12 + 0;
		indexes[starIndex * 30 + 25] = starIndex * 12 + 6;
		indexes[starIndex * 30 + 26] = starIndex * 12 + 8;
		indexes[starIndex * 30 + 27] = starIndex * 12 + 0;
		indexes[starIndex * 30 + 28] = starIndex * 12 + 8;
		indexes[starIndex * 30 + 29] = starIndex * 12 + 10;
	}
	
	glGenBuffers(1, &self->vboID);
	glGenBuffers(1, &self->eboID);
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_c4ub) * 84, NULL, GL_STATIC_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * 210, indexes, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void SparkleEffect_dispose(compat_type(SparkleEffect *) selfPtr) {
	SparkleEffect * self = selfPtr;
	
	if (self->owner != NULL) {
		self->owner->eventDispatcher->unregisterForEvent(self->owner->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
	}
	glDeleteBuffers(1, &self->vboID);
	glDeleteBuffers(1, &self->eboID);
	GameObjectView_dispose(self);
}

void SparkleEffect_draw(compat_type(SparkleEffect *) selfPtr, Matrix modelviewMatrix, double deltaTime) {
	SparkleEffect * self = selfPtr;
	struct vertex_p2f_c4ub vertex, * vertices;
	int starIndex, pointIndex;
	float step;
	float minRadius, maxRadius;
	const float starAngles[5] = {
		M_PI * 0.28125f,
		M_PI * 0.6875f,
		M_PI * 1.125f,
		M_PI * 1.4375f,
		M_PI * 1.78125f
	};
	Vector2 starOffset;
	
	if (self->owner != NULL) {
		self->position = self->owner->position;
	}
	
	self->currentTime += deltaTime;
	if (self->currentTime >= DURATION) {
		self->markedForRemoval = true;
		step = 1.0f;
	} else {
		step = pow(self->currentTime / DURATION, 0.5f);
	}
	minRadius = (1.0f - step) * (0.25f + step * 0.5f) * self->radius;
	maxRadius = (1.0f + step) * (0.25f + step * 0.375f) * self->radius;
	
	Matrix_translate(&modelviewMatrix, self->position.x, self->position.y, 0.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	vertices = glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
	
	vertex.color[0] = 0xFF;
	vertex.color[1] = 0xFF;
	vertex.color[2] = 0xFF;
	vertex.color[3] = 0xFF;
	for (starIndex = 0; starIndex < 5; starIndex++) {
		starOffset.x = cos(starAngles[starIndex]) * STAR_OFFSET * pow(step, 0.75f) * self->radius * RATIO;
		starOffset.y = sin(starAngles[starIndex]) * STAR_OFFSET * pow(step, 0.75f) * self->radius;
		for (pointIndex = 0; pointIndex < 6; pointIndex++) {
			vertex.position[0] = starOffset.x + cos(M_PI * 2 * (pointIndex * 2 - 1) / 12) * minRadius * STAR_RADIUS_0 * RATIO;
			vertex.position[1] = starOffset.y + sin(M_PI * 2 * (pointIndex * 2 - 1) / 12) * minRadius * STAR_RADIUS_0;
			vertices[starIndex * 12 + pointIndex * 2 + 0] = vertex;
			vertex.position[0] = starOffset.x + cos(M_PI * 2 * pointIndex / 6) * maxRadius * STAR_RADIUS_0 * RATIO;
			vertex.position[1] = starOffset.y + sin(M_PI * 2 * pointIndex / 6) * maxRadius * STAR_RADIUS_0;
			vertices[starIndex * 12 + pointIndex * 2 + 1] = vertex;
		}
	}
	for (pointIndex = 0; pointIndex < 6; pointIndex++) {
		vertex.position[0] = cos(M_PI * 2 * (pointIndex * 2 - 1) / 12) * minRadius * STAR_RADIUS_1 * RATIO;
		vertex.position[1] = sin(M_PI * 2 * (pointIndex * 2 - 1) / 12) * minRadius * STAR_RADIUS_1;
		vertices[starIndex * 12 + pointIndex * 2 + 0] = vertex;
		vertex.position[0] = cos(M_PI * 2 * pointIndex / 6) * maxRadius * STAR_RADIUS_1 * RATIO;
		vertex.position[1] = sin(M_PI * 2 * pointIndex / 6) * maxRadius * STAR_RADIUS_1;
		vertices[starIndex * 12 + pointIndex * 2 + 1] = vertex;
	}
	starIndex++;
	vertex.color[0] = 0xEC;
	vertex.color[1] = 0x63;
	vertex.color[2] = 0x98;
	vertex.color[3] = 0xFF;
	for (pointIndex = 0; pointIndex < 6; pointIndex++) {
		vertex.position[0] = cos(M_PI * 2 * pointIndex / 6) * minRadius * STAR_RADIUS_2 * RATIO;
		vertex.position[1] = sin(M_PI * 2 * pointIndex / 6) * minRadius * STAR_RADIUS_2;
		vertices[starIndex * 12 + pointIndex * 2 + 0] = vertex;
		vertex.position[0] = cos(M_PI * 2 * (pointIndex * 2 + 1) / 12) * maxRadius * STAR_RADIUS_2 * RATIO;
		vertex.position[1] = sin(M_PI * 2 * (pointIndex * 2 + 1) / 12) * maxRadius * STAR_RADIUS_2;
		vertices[starIndex * 12 + pointIndex * 2 + 1] = vertex;
	}
	
	glUnmapBuffer(GL_ARRAY_BUFFER);
	
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_c4ub), (void *) offsetof(struct vertex_p2f_c4ub, position));
	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(struct vertex_p2f_c4ub), (void *) offsetof(struct vertex_p2f_c4ub, color));
	glDrawElements(GL_TRIANGLES, 210, GL_UNSIGNED_SHORT, 0);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
