/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/EnemyModel.h"
#include "convergence/PlayerModel.h"
#include "convergence/SwordSwingModel.h"
#include <math.h>
#include <stdlib.h>

SwordSwingModel * SwordSwingModel_create(DungeonRoomStateModel * stateModel, compat_type(MovingEntityModel *) entity, bool hostile, int damage, float knockback, float minRadius, float maxRadius, float ratio, int length) {
	stemobject_create_implementation(SwordSwingModel, init, stateModel, entity, hostile, damage, knockback, minRadius, maxRadius, ratio, length)
}

static void sharedInit(SwordSwingModel * self) {
	self->dispose = SwordSwingModel_dispose;
	self->copy = SwordSwingModel_copy;
	self->update = SwordSwingModel_update;
	self->atRest = SwordSwingModel_atRest;
	self->roughBounds = SwordSwingModel_roughBounds;
	self->collisionGeometry = SwordSwingModel_collisionGeometry;
	self->wouldCollide = SwordSwingModel_wouldCollide;
	
	self->type = GAME_OBJECT_TYPE_SWORD_SWING;
}

static bool ownerDied(void * sender, Atom eventID, void * eventData, void * context) {
	SwordSwingModel * self = context;
	
	self->markedForRemoval = true;
	self->entity->eventDispatcher->unregisterForEvent(self->entity->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), ownerDied, self);
	self->entity->eventDispatcher->unregisterForEvent(self->entity->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), ownerDied, self);
	self->entity = NULL;
	return true;
}

void SwordSwingModel_init(compat_type(SwordSwingModel *) selfPtr, DungeonRoomStateModel * stateModel, compat_type(MovingEntityModel *) entity, bool hostile, int damage, float knockback, float minRadius, float maxRadius, float ratio, int length) {
	SwordSwingModel * self = selfPtr;
	
	GameObjectModel_init(self, stateModel);
	sharedInit(self);
	
	self->rawProgress = 0;
	self->progress = 0.0f;
	self->entity = entity;
	self->entity->eventDispatcher->registerForEvent(self->entity->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), ownerDied, self);
	self->entity->eventDispatcher->registerForEvent(self->entity->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), ownerDied, self);
	self->position = self->entity->position;
	self->facing = self->entity->facing;
	self->hostile = hostile;
	self->damage = damage;
	self->knockback = knockback;
	self->minRadius = minRadius;
	self->maxRadius = maxRadius;
	self->ratio = ratio;
	self->length = length;
}

void SwordSwingModel_initCopy(compat_type(SwordSwingModel *) selfPtr, compat_type(SwordSwingModel *) originalPtr) {
	SwordSwingModel * self = selfPtr, * original = originalPtr;
	
	GameObjectModel_initCopy(self, original);
	sharedInit(self);
	
	self->rawProgress = original->rawProgress;
	self->entity = original->entity;
	if (self->entity != NULL) {
		self->entity->eventDispatcher->registerForEvent(self->entity->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), ownerDied, self);
		self->entity->eventDispatcher->registerForEvent(self->entity->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), ownerDied, self);
	}
	self->progress = original->progress;
	self->position = original->position;
	self->facing = original->facing;
	self->hostile = original->hostile;
	self->damage = original->damage;
	self->knockback = original->knockback;
	self->minRadius = original->minRadius;
	self->maxRadius = original->maxRadius;
	self->ratio = original->ratio;
	self->length = original->length;
}

void SwordSwingModel_dispose(compat_type(SwordSwingModel *) selfPtr) {
	SwordSwingModel * self = selfPtr;
	
	if (self->entity != NULL) {
		self->entity->eventDispatcher->unregisterForEvent(self->entity->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), ownerDied, self);
		self->entity->eventDispatcher->unregisterForEvent(self->entity->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), ownerDied, self);
		self->entity = NULL;
	}
	GameObjectModel_dispose(self);
}

compat_type(SwordSwingModel *) SwordSwingModel_copy(compat_type(SwordSwingModel *) selfPtr) {
	stemobject_copy_implementation(SwordSwingModel, initCopy)
}

void SwordSwingModel_update(compat_type(SwordSwingModel *) selfPtr) {
	SwordSwingModel * self = selfPtr;
	
	self->rawProgress++;
	self->progress = (-cos(self->rawProgress * M_PI / self->length) + 1.0f) * 0.5f;
	if (self->rawProgress > self->length) {
		self->markedForRemoval = true;
	}
}

bool SwordSwingModel_atRest(compat_type(SwordSwingModel *) selfPtr) {
	return false;
}

struct rect4f SwordSwingModel_roughBounds(compat_type(SwordSwingModel *) selfPtr) {
	SwordSwingModel * self = selfPtr;
	struct rect4f bounds;
	float angle;
	Vector2 startPoint, midPoint, endPoint;
	
	if (self->entity != NULL) {
		self->position = self->entity->position;
		self->facing = self->entity->facing;
	}
	
	angle = directionToRadians(self->facing);
	startPoint.x = self->maxRadius * self->ratio *  cos(angle);
	startPoint.y = self->maxRadius * self->ratio * -sin(angle);
	midPoint.x = self->maxRadius * sin(angle);
	midPoint.y = self->maxRadius * cos(angle);
	endPoint.x = -self->maxRadius * self->ratio *  cos(angle);
	endPoint.y = -self->maxRadius * self->ratio * -sin(angle);
	
	bounds.left = self->position.x + min(min(min(startPoint.x, endPoint.x), startPoint.x + midPoint.x), endPoint.x + midPoint.x);
	bounds.right = self->position.x + max(max(max(startPoint.x, endPoint.x), startPoint.x + midPoint.x), endPoint.x + midPoint.x);
	bounds.bottom = self->position.y + min(min(min(startPoint.y, endPoint.y), startPoint.y + midPoint.y), endPoint.y + midPoint.y);
	bounds.top = self->position.y + max(max(max(startPoint.y, endPoint.y), startPoint.y + midPoint.y), endPoint.y + midPoint.y);
	
	return bounds;
}

struct collisionGeometry SwordSwingModel_collisionGeometry(compat_type(SwordSwingModel *) selfPtr) {
	SwordSwingModel * self = selfPtr;
	struct collisionGeometry geometry;
	float angle;
	
	if (self->entity != NULL) {
		self->position = self->entity->position;
		self->facing = self->entity->facing;
	}
	
	geometry.type = COLLISION_GEOMETRY_TYPE_LINE;
	angle = directionToRadians(self->facing);
	geometry.u.line.endpoints[0].x = self->position.x + cos(self->progress * M_PI) * self->minRadius *  cos(angle) + sin(self->progress * M_PI) * self->minRadius * sin(angle);
	geometry.u.line.endpoints[0].y = self->position.y + cos(self->progress * M_PI) * self->minRadius * -sin(angle) + sin(self->progress * M_PI) * self->minRadius * cos(angle);
	geometry.u.line.endpoints[1].x = self->position.x + cos(self->progress * M_PI) * self->maxRadius * self->ratio *  cos(angle) + sin(self->progress * M_PI) * self->maxRadius * sin(angle);
	geometry.u.line.endpoints[1].y = self->position.y + cos(self->progress * M_PI) * self->maxRadius * self->ratio * -sin(angle) + sin(self->progress * M_PI) * self->maxRadius * cos(angle);
	geometry.u.line.effectiveNormal = directionToVector(self->facing);
	return geometry;
}

bool SwordSwingModel_wouldCollide(compat_type(SwordSwingModel *) selfPtr, compat_type(GameObjectModel *) objectPtr) {
	SwordSwingModel * self = selfPtr;
	
	switch (((GameObjectModel *) objectPtr)->type) {
		case GAME_OBJECT_TYPE_ENEMY:
			return !self->hostile;
		case GAME_OBJECT_TYPE_PLAYER:
			return self->hostile;
		default:
			return false;
	}
}
