/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/AudioManager.h"
#include "convergence/ChooseClassState.h"
#include "convergence/ConfigureControlsState.h"
#include "convergence/GameplayState.h"
#include "convergence/GameSession.h"
#include "convergence/InstructionsState.h"
#include "convergence/PreferencesState.h"
#include "convergence/SharedEvents.h"
#include "convergence/ShellStateGlobals.h"
#include "convergence/SplashScreenState.h"
#include "gamepad/Gamepad.h"
#include "glgraphics/GLIncludes.h"
#include "shell/Shell.h"
#include "shell/ShellBatteryInfo.h"
#include "shell/ShellKeyCodes.h"
#include "shell/Target.h"
#include "statecontroller/StateController.h"
#include "utilities/Random.h"

#if defined(STEM_PLATFORM_macosx)
#include "nsopenglshell/NSOpenGLTarget.h"
#else
#include "glutshell/GLUTTarget.h"
#endif

#include <stdio.h>
#include <string.h>
#include <time.h>

static StateController * stateController;
static GameSession * gameSession;
static const char * recordInputToFile;
static const char * playInputFromFile;

void Target_draw() {
	stateController->currentState->draw(stateController->currentState);
}

void Target_keyDown(unsigned int charCode, unsigned int keyCode, unsigned int modifiers) {
	struct keyEvent event;
	
	if (keyCode == KEYBOARD_RETURN_OR_ENTER && (modifiers & MODIFIER_ALT_BIT)) {
		Shell_setFullScreen(!Shell_isFullScreen());
		stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(Shell_isFullScreen() ? EVENT_FULLSCREENED : EVENT_WINDOWED), NULL);
	} else {
		event.keyCode = keyCode;
		event.charCode = charCode;
		event.modifiers = modifiers;
		stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), &event);
	}
}

void Target_keyUp(unsigned int charCode, unsigned int keyCode, unsigned int modifiers) {
	struct keyEvent event;
	
	event.keyCode = keyCode;
	event.charCode = charCode;
	event.modifiers = modifiers;
	stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_KEY_UP), &event);
}

void Target_keyModifiersChanged(unsigned int modifiers) {
	struct keyEvent event;
	
	event.modifiers = modifiers;
	stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_KEY_MODIFIERS_CHANGED), &event);
}

void Target_mouseDown(unsigned int buttonNumber, float x, float y) {
	struct mouseEvent event;
	
	event.button = buttonNumber;
	event.x = x;
	event.y = y;
	stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), &event);
}

void Target_mouseUp(unsigned int buttonNumber, float x, float y) {
	struct mouseEvent event;
	
	event.button = buttonNumber;
	event.x = x;
	event.y = y;
	stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_UP), &event);
}

void Target_mouseMoved(float x, float y) {
	struct mouseEvent event;
	
	event.x = x;
	event.y = y;
	stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_MOVED), &event);
}

void Target_mouseDragged(unsigned int buttonMask, float x, float y) {
	struct mouseEvent event;
	
	event.button = buttonMask;
	event.x = x;
	event.y = y;
	stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DRAGGED), &event);
}

void Target_resized(unsigned int newWidth, unsigned int newHeight) {
	g_windowWidth = newWidth;
	g_windowHeight = newHeight;
	glViewport(0, 0, newWidth, newHeight);
	if (stateController != NULL) {
		stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_RESIZED), NULL);
	}
}

void Target_backgrounded() {
	if (stateController != NULL) {
		stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_BACKGROUNDED), NULL);
	}
}

void Target_foregrounded() {
	if (stateController != NULL) {
		stateController->eventDispatcher->dispatchEvent(stateController->eventDispatcher, Atom_fromString(EVENT_FOREGROUNDED), NULL);
	}
}

#if defined(STEM_PLATFORM_macosx)

void NSOpenGLTarget_configure(int argc, const char ** argv, struct NSOpenGLShellConfiguration * configuration) {
	unsigned int screenWidth, screenHeight;
	int argIndex;
	
	for (argIndex = 0; argIndex < argc; argIndex++) {
		if (!strcmp(argv[argIndex], "-record") && argIndex < argc - 1) {
			recordInputToFile = argv[++argIndex];
			printf("Input will be recorded to \"%s\"\n", recordInputToFile);
		} else if (!strcmp(argv[argIndex], "-playback") && argIndex < argc - 1) {
			playInputFromFile = argv[++argIndex];
			printf("Input will be played back from \"%s\"\n", playInputFromFile);
		}
	}
	
	NSOpenGLShell_getMainScreenDimensions(&screenWidth, &screenHeight);
	configuration->windowWidth = 1024;
	configuration->windowHeight = 768;
	configuration->windowX = (screenWidth - configuration->windowWidth) / 2;
	configuration->windowY = screenHeight - (screenHeight - configuration->windowHeight) / 2;
	configuration->windowTitle = "Convergence";
	
	g_windowWidth = configuration->windowWidth;
	g_windowHeight = configuration->windowHeight;
}

#else

void GLUTTarget_configure(int argc, const char ** argv, struct GLUTShellConfiguration * configuration) {
	int argIndex;
	
	for (argIndex = 0; argIndex < argc; argIndex++) {
		if (!strcmp(argv[argIndex], "-record") && argIndex < argc - 1) {
			recordInputToFile = argv[++argIndex];
			printf("Input will be recorded to \"%s\"\n", recordInputToFile);
		} else if (!strcmp(argv[argIndex], "-playback") && argIndex < argc - 1) {
			playInputFromFile = argv[++argIndex];
			printf("Input will be played back from \"%s\"\n", playInputFromFile);
		}
	}
	
	g_windowWidth = configuration->windowWidth;
	g_windowHeight = configuration->windowHeight;
	
	configuration->windowTitle = "Convergence";
}

#endif

static void shutdownAudio() {
	AudioManager_globalShutdown();
}

void Target_init() {
	SplashScreenState * splashScreenState;
	ChooseClassState * chooseClassState;
	InstructionsState * instructionsState;
	PreferencesState * preferencesState;
	ConfigureControlsState * configureControlsState;
	GameplayState * gameplayState;
	unsigned int randomDiscardIndex;
	
	sdrand(time(NULL));
	for (randomDiscardIndex = 0; randomDiscardIndex < 50; randomDiscardIndex++) {
		irand();
	}
	
	gameSession = GameSession_create();
	gameSession->recordInputToFile = recordInputToFile;
	gameSession->playInputFromFile = playInputFromFile;
	
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture", "textures/apple.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture", "textures/cloud.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture", "textures/gem.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture", "textures/rainbow.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture", "textures/staircase.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture_canvas", "textures/tilecanvas_floor0.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture_canvas", "textures/tilecanvas_floor1.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "texture_canvas", "textures/tilecanvas_floor2.json");
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "bitmapfont", "font.json");
	
	AudioManager_globalInit(gameSession);
	atexit(shutdownAudio);
	
	Gamepad_init();
	
	stateController = StateController_create();
	splashScreenState = SplashScreenState_create(gameSession);
	chooseClassState = ChooseClassState_create(gameSession);
	instructionsState = InstructionsState_create(gameSession);
	preferencesState = PreferencesState_create(gameSession);
	configureControlsState = ConfigureControlsState_create(gameSession);
	gameplayState = GameplayState_create(gameSession);
	stateController->addState(stateController, splashScreenState);
	stateController->addState(stateController, chooseClassState);
	stateController->addState(stateController, instructionsState);
	stateController->addState(stateController, preferencesState);
	stateController->addState(stateController, configureControlsState);
	stateController->addState(stateController, gameplayState);
	stateController->addTransition(stateController, splashScreenState, chooseClassState, "choose_class");
	stateController->addTransition(stateController, splashScreenState, instructionsState, "instructions");
	stateController->addTransition(stateController, splashScreenState, preferencesState, "preferences");
	stateController->addTransition(stateController, chooseClassState, gameplayState, "start_game");
	stateController->addTransition(stateController, chooseClassState, splashScreenState, "back");
	stateController->addTransition(stateController, chooseClassState, instructionsState, "instructions");
	stateController->addTransition(stateController, instructionsState, splashScreenState, "back");
	stateController->addTransition(stateController, instructionsState, gameplayState, "start_game");
	stateController->addTransition(stateController, gameplayState, splashScreenState, "game_over");
	stateController->addTransition(stateController, preferencesState, configureControlsState, "configure_controls");
	stateController->addTransition(stateController, preferencesState, splashScreenState, "done");
	stateController->addTransition(stateController, configureControlsState, preferencesState, "done");
	stateController->setState(stateController, splashScreenState);
	
	gameSession->resourceManager->referenceResource(gameSession->resourceManager, "bitmapfont", "font.json");
	
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	
	Shell_setFullScreen(gameSession->preferences->getBoolean(gameSession->preferences, "fullscreen"));
	
	Shell_mainLoop();
}
